# cython: language_level=3

# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

from numpy import array, empty, int8, float64, concatenate, diff
from numpy.linalg import solve

cimport cython
from numpy cimport ndarray, npy_intp
from cherab.core.math.interpolators.utility cimport find_index, lerp, derivatives_array, factorial
from libc.math cimport INFINITY, NAN

# internal constants used to represent the different extrapolation options
DEF EXT_NEAREST = 0
DEF EXT_LINEAR = 1
DEF EXT_QUADRATIC = 2

# map extrapolation string to relevant internal constant
_EXTRAPOLATION_TYPES = {
    'nearest': EXT_NEAREST,
    'linear': EXT_LINEAR,
    'quadratic': EXT_QUADRATIC
}

cdef class _Interpolate2DBase(Function2D):
    """
    Base class for 2D interpolators. Coordinates and data arrays are here
    sorted and transformed into numpy arrays.

    :param object x: An array-like object containing real values.
    :param object y: An array-like object containing real values.
    :param object f: A 2D array-like object of sample values corresponding to the
    `x` and `y` array points.
    :param bint extrapolate: optional
    If True, the extrapolation of data is enabled outside the range of the
    data set. The default is False. A ValueError is raised if extrapolation
    is disabled and a point is requested outside the data set.
    :param object extrapolation_type: optional
    Sets the method of extrapolation. The options are: 'nearest' (default)
    and other options given by the subclass.
    :param double extrapolation_range: optional
    The attribute can be set to limit the range beyond the data set bounds
    that extrapolation is permitted. The default range is set to infinity.
    Requesting data beyond the extrapolation range will result in a
    ValueError being raised.
    :param tolerate_single_value: optional
    If True, single-value arrays will be tolerated as inputs. If a single
    value is supplied, that value will be extrapolated over the entire
    real range. If False (default), supplying a single value will result
    in a ValueError being raised.
    """

    def __init__(self, object x, object y, object f, bint extrapolate=False, str extrapolation_type='nearest',
                 double extrapolation_range=INFINITY, bint tolerate_single_value=False):

        # convert data to c-contiguous numpy arrays
        x = array(x, dtype=float64, order='c')
        y = array(y, dtype=float64, order='c')
        f = array(f, dtype=float64, order='c')

        # check dimensions are 1D
        if x.ndim != 1:
            raise ValueError("The x array must be 1D.")

        if y.ndim != 1:
            raise ValueError("The y array must be 1D.")

        # check data is 2D
        if f.ndim != 2:
            raise ValueError("The f array must be 2D.")

        # check the shapes of data and coordinates are consistent
        shape = (x.shape[0], y.shape[0])
        if f.shape != shape:
            raise ValueError("The dimension and data arrays must have consistent shapes ((x, y)={}, f={}).".format(shape, f.shape))

        # check the dimension arrays must be monotonically increasing
        if (diff(x) <= 0).any():
            raise ValueError("The x array must be monotonically increasing.")

        if (diff(y) <= 0).any():
            raise ValueError("The y array must be monotonically increasing.")

        # extrapolation is controlled internally by setting a positive extrapolation_range
        if extrapolate:
            self._extrapolation_range = max(0, extrapolation_range)
        else:
            self._extrapolation_range = 0

        # map extrapolation type name to internal constant
        if extrapolation_type in _EXTRAPOLATION_TYPES:
            self._extrapolation_type = _EXTRAPOLATION_TYPES[extrapolation_type]
        else:
            raise ValueError("Extrapolation type {} does not exist.".format(extrapolation_type))

        # populate internal arrays and memory views
        self._x = x
        self._y = y

        # Check for single value in x input
        if x.shape[0] == 1 and not tolerate_single_value:
            raise ValueError("There is only a single value in the x array. "
                             "Consider turning on the 'tolerate_single_value' argument.")

        # Check for single value in y input
        if y.shape[0] == 1 and not tolerate_single_value:
            raise ValueError("There is only a single value in the y array. "
                             "Consider turning on the 'tolerate_single_value' argument.")

        # build internal state of interpolator
        self._build(x, y, f)

    cdef object _build(self, ndarray x, ndarray y, ndarray f):
        """
        Build additional internal state.
        
        Implement in sub-classes that require additional state to be build
        from the source arrays.        
            
        :param x: x array 
        :param y: y array
        :param f: f array 
        """
        pass

    cdef double evaluate(self, double px, double py) except? -1e999:
        """
        Evaluate the interpolating function.

        :param double px, double py: coordinates
        :return: the interpolated value
        """

        return self._dispatch(px, py, order_x=0, order_y=0)

    cpdef double derivative(self, double px, double py, int order_x, int order_y) except? -1e999:
        """
        Evaluate the interpolating function.

        :param double px, double py: coordinates
        :return: the interpolated value
        """

        if order_x < 1 and order_y < 1:
            raise ValueError('At least one derivative order must be > 0.')

        if order_x < 0:
            raise ValueError('The y derivative order cannot be less than zero.')

        if order_y < 0:
            raise ValueError('The y derivative order cannot be less than zero.')

        return self._dispatch(px, py, order_x, order_y)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef double _dispatch(self, double px, double py, int order_x, int order_y) except? -1e999:
        """
        Identifies the region the sample point lies in and calls the relevant evaluator.
        """

        cdef:
            int ix, iy, nx, ny
            double[::1] x, y

        x = self._x
        y = self._y

        nx = x.shape[0]
        ny = y.shape[0]

        ix = find_index(x, px, self._extrapolation_range)
        iy = find_index(y, py, self._extrapolation_range)

        if 0 <= ix < nx - 1:
            if 0 <= iy < ny - 1:
                return self._evaluate(px, py, order_x, order_y, ix, iy)
            elif iy == -1:
                return self._extrapolate(px, py, order_x, order_y, ix, 0, px, y[0], True, False)
            elif iy == ny - 1:
                return self._extrapolate(px, py, order_x, order_y, ix, ny - 2, px, y[ny - 1], True, False)

        elif ix == -1:
            if 0 <= iy < ny - 1:
                return self._extrapolate(px, py, order_x, order_y, 0, iy, x[0], py, False, True)
            elif iy == -1:
                return self._extrapolate(px, py, order_x, order_y, 0, 0, x[0], y[0], False, False)
            elif iy == ny - 1:
                return self._extrapolate(px, py, order_x, order_y, 0, ny - 2, x[0], y[ny - 1], False, False)

        elif ix == nx - 1:
            if 0 <= iy < ny - 1:
                return self._extrapolate(px, py, order_x, order_y, nx - 2, iy, x[nx - 1], py, False, True)
            elif iy == -1:
                return self._extrapolate(px, py, order_x, order_y, nx - 2, 0, x[nx - 1], y[0], False, False)
            elif iy == ny - 1:
                return self._extrapolate(px, py, order_x, order_y, nx - 2, ny - 2, x[nx - 1], y[ny - 1], False, False)

        # value is outside of permitted limits
        min_range_x = x[0] - self._extrapolation_range
        max_range_x = x[nx - 1] + self._extrapolation_range

        min_range_y = y[0] - self._extrapolation_range
        max_range_y = y[ny - 1] + self._extrapolation_range

        raise ValueError("The specified value (x={}, y={}) is outside the range of the supplied data and/or extrapolation range: "
                         "x bounds=({}, {}), y bounds=({}, {})".format(px, py, min_range_x, max_range_x, min_range_y, max_range_y))

    cdef double _evaluate(self, double px, double py, int order_x, int order_y, int ix, int iy) except? -1e999:
        """
        Evaluate the interpolating function which is valid in the area given
        by 'ix' and 'iy' at any position ('px', 'py').

        :param double px, double py: coordinates
        :param int ix, int iy: indices of the area of interest
        :return: the interpolated value
        """
        raise NotImplementedError("This abstract method has not been implemented yet.")

    cdef double _extrapolate(self, double px, double py, int order_x, int order_y, int ix, int iy,
                             double nearest_px, double nearest_py, bint inside_x, bint inside_y) except? -1e999:
        """
        Extrapolate the interpolation function valid on area given by
        'ix' and 'iy' to position ('px', 'py').
        
        :param px: Requested x coord.
        :param py: Requested y coord.
        :param order_x: Derivative order along x.
        :param order_y: Derivative order along y.
        :param ix: Lower x index of closest array bin.
        :param iy: Lower y index of closest array bin.
        :param nearest_px: Closest x coordinate inside interpolation domain.
        :param nearest_py: Closest y coordinate inside interpolation domain. 
        :param inside_x: If the x coord is inside the interpolation domain.
        :param inside_y: If the y coord is inside the interpolation domain.
        :return: The extrapolated value.
        """

        if self._extrapolation_type == EXT_NEAREST:
            return self._extrapol_nearest(px, py, order_x, order_y, ix, iy, nearest_px, nearest_py, inside_x, inside_y)

        elif self._extrapolation_type == EXT_LINEAR:
            return self._extrapol_linear(px, py, order_x, order_y, ix, iy, nearest_px, nearest_py, inside_x, inside_y)

        elif self._extrapolation_type == EXT_QUADRATIC:
            return self._extrapol_quadratic(px, py, order_x, order_y, ix, iy, nearest_px, nearest_py, inside_x, inside_y)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef double _extrapol_nearest(self, double px, double py, int order_x, int order_y, int ix, int iy, double nearest_px, double nearest_py, bint inside_x, bint inside_y) except? -1e999:
        """
        Extrapolate using the nearest point of the domain.
        """

        cdef:
            int nx, ny
            double[::1] x, y

        x = self._x
        y = self._y

        nx = x.shape[0]
        ny = y.shape[0]

        # f(x,y)
        if order_x == 0 and order_y == 0:
            return self._evaluate(nearest_px, nearest_py, 0, 0, ix, iy)

        # derivatives
        # there are 8 separate domains - the 4 corners (that are flat) and the 4
        # mid sections that mirror the profile of the closest edge of the domain
        if inside_x:

            if order_y > 0:
                return 0.0
            return self._evaluate(nearest_px, nearest_py, order_x, 0, ix, iy)

        elif inside_y:

            if order_x > 0:
                return 0.0
            return self._evaluate(nearest_px, nearest_py, 0, order_y, ix, iy)

        else:

            # corner
            return 0.0

    cdef double _extrapol_linear(self, double px, double py, int order_x, int order_y, int ix, int iy, double nearest_px, double nearest_py, bint inside_x, bint inside_y) except? -1e999:
        """
        Extrapolate linearly the interpolation function valid on area given by
        'ix' and 'iy' to position ('px', 'py').

        :param double px, double py: coordinates
        :param int ix, int iy: indices of the area of interest
        :param double nearest_px, nearest_py: the nearest position from
        ('px', 'py') in the interpolation domain.
        :return: the extrapolated value
        """
        raise NotImplementedError("There is no linear extrapolation available for this interpolation.")

    cdef double _extrapol_quadratic(self, double px, double py, int order_x, int order_y, int ix, int iy, double nearest_px, double nearest_py, bint inside_x, bint inside_y) except? -1e999:
        """
        Extrapolate quadratically the interpolation function valid on area given by
        'ix' and 'iy' to position ('px', 'py').

        :param double px, double py: coordinates
        :param int ix, int iy: indices of the area of interest
        :param double nearest_px, nearest_py: the nearest position from
        ('px', 'py') in the interpolation domain.
        :return: the extrapolated value
        """
        raise NotImplementedError("There is no quadratic extrapolation available for this interpolation.")


cdef class Interpolate2DLinear(_Interpolate2DBase):
    """
    .. deprecated:: 1.3.0
       Use `raysect.core.math.function.float.Interpolator2DArray` instead.

    Interpolates 2D data using linear interpolation.

    Inherits from Function2D, implements `__call__(x, y)`.

    :param object x: An array-like object containing real values.
    :param object y: An array-like object containing real values.
    :param object f: A 2D array-like object of sample values corresponding to the
      `x` and `y` array points.
    :param bint extrapolate: If True, the extrapolation of data is enabled outside
      the range of the data set. The default is False. A ValueError is raised if
      extrapolation is disabled and a point is requested outside the data set.
    :param object extrapolation_type: Sets the method of extrapolation. The options
      are: 'nearest' (default), 'linear'
    :param double extrapolation_range: The attribute can be set to limit the range
      beyond the data set bounds that extrapolation is permitted. The default range
      is set to infinity. Requesting data beyond the extrapolation range will result
      in a ValueError being raised.
    :param tolerate_single_value: If True, single-value arrays will be tolerated as
      inputs. If a single value is supplied, that value will be extrapolated over
      the entire real range. If False (default), supplying a single value will
      result in a ValueError being raised.

    .. code-block:: pycon

       >>> import numpy as np
       >>> from cherab.core.math import Interpolate2DLinear
       >>>
       >>> # implements x**2 + y
       >>> drange = np.linspace(-2.5, 2.5, 100)
       >>> values = np.zeros((100, 100))
       >>> for i in range(100):
       >>>     for j in range(100):
       >>>         values[i, j] = drange[i]**2 + drange[j]
       >>>
       >>> f2d = Interpolate2DLinear(drange, drange, values)
       >>>
       >>> f2d(0, 0)
       0.00063769
       >>> f2d(-2, 1)
       5.00022956
       >>> f2d(-2, 3)
       ValueError: The specified value (x=-2.0, y=3.0) is outside the range of the supplied
       data and/or extrapolation range: x bounds=(-2.5, 2.5), y bounds=(-2.5, 2.5)
    """

    def __init__(self, object x, object y, object f, bint extrapolate=False, str extrapolation_type='nearest',
                 double extrapolation_range=float('inf'), bint tolerate_single_value=False):

        supported_extrapolations = ['nearest', 'linear']

        extrapolation_type = extrapolation_type.lower()
        if extrapolation_type not in supported_extrapolations:
            raise ValueError("Unsupported extrapolation type: {}".format(extrapolation_type))

        super().__init__(x, y, f, extrapolate, extrapolation_type, extrapolation_range, tolerate_single_value)

    cdef object _build(self, ndarray x, ndarray y, ndarray f):

        cdef ndarray temp

        # if x array is single valued, expand x array and data along x axis to simplify interpolation
        if x.shape[0] == 1:

            # set x array to full real range
            self._x = array([-INFINITY, INFINITY], dtype=float64)

            # duplicate data to provide a pseudo range for interpolation coefficient calculation
            temp = f
            x = array([-1., +1.], dtype=float64)
            f = empty((2, f.shape[1]), dtype=float64)
            f[:, :] = temp

        # if y array is single valued, expand y array and data along y axis to simplify interpolation
        if y.shape[0] == 1:

            # set y array to full real range
            self._y = array([-INFINITY, INFINITY], dtype=float64)

            # duplicate data to provide a pseudo range for interpolation coefficient calculation
            temp = f
            y = array([-1., +1.], dtype=float64)
            f = empty((f.shape[0], 2), dtype=float64)
            f[:, :] = temp

        # create memory views
        self._wx = x
        self._wy = y
        self._wf = f

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    @cython.cdivision(True)
    cdef double _evaluate(self, double px, double py, int order_x, int order_y, int ix, int iy) except? -1e999:
        """
        Evaluate the interpolating function which is valid in the area given
        by 'ix' and 'iy' at any position ('px', 'py').

        :param double px, double py: coordinates
        :param int ix, int iy: indices of the area of interest
        :return: the interpolated value
        """
        cdef:
            double[::1] x, y
            double[:,::1] f
            double t0, t1

        x = self._wx
        y = self._wy
        f = self._wf

        if order_x == 0:

            # f(x, y)
            if order_y == 0:

                t0 = lerp(x[ix], x[ix+1], f[ix, iy],   f[ix+1, iy],   px)
                t1 = lerp(x[ix], x[ix+1], f[ix, iy+1], f[ix+1, iy+1], px)
                return lerp(y[iy], y[iy+1], t0, t1, py)

            # df(x, y) / dy
            elif order_y == 1:

                t0 = (f[ix, iy+1] - f[ix, iy]) / (y[iy+1] - y[iy])
                t1 = (f[ix+1, iy+1] - f[ix+1, iy]) / (y[iy+1] - y[iy])
                return lerp(x[ix], x[ix+1], t0, t1, px)

        elif order_x == 1:

            # df(x, y) / dx
            if order_y == 0:

                t0 = (f[ix+1, iy] - f[ix, iy]) / (x[ix+1] - x[ix])
                t1 = (f[ix+1, iy+1] - f[ix, iy+1]) / (x[ix+1] - x[ix])
                return lerp(y[iy], y[iy+1], t0, t1, py)

            # d2f(x, y) / dxdy
            elif order_y == 1:

                t0 = self._evaluate(px, y[iy], 1, 0, ix, iy)
                t1 = self._evaluate(px, y[iy+1], 1, 0, ix, iy)
                return (t1 - t0) / (y[iy+1] - y[iy])

        # higher order derivatives
        return 0.0

    cdef double _extrapol_linear(self, double px, double py, int order_x, int order_y, int ix, int iy, double nearest_px, double nearest_py, bint inside_x, bint inside_y) except? -1e999:
        """
        Extrapolate linearly the interpolation function valid on area given by
        'ix' and 'iy' to position ('px', 'py').

        :param double px, double py: coordinates
        :param int ix, int iy: indices of the area of interest
        :param double nearest_px, nearest_py: the nearest position from
        ('px', 'py') in the interpolation domain.
        :return: the extrapolated value
        """

        return self._evaluate(px, py, order_x, order_y, ix, iy)


cdef class Interpolate2DCubic(_Interpolate2DBase):
    """
    .. deprecated:: 1.3.0
       Use `raysect.core.math.function.float.Interpolator2DArray` instead.

    Interpolates 2D data using cubic interpolation.

    Inherits from Function2D, implements `__call__(x, y)`.

    Data and coordinates are first normalised to the range [0, 1] so as to
    prevent inaccuracy from float numbers. A local calculation based on finite
    differences is used. The splines coefficients are not calculated before
    evaluation but on demand only and are cached as they are calculated. Plus,
    only one polynomial is calculated at each evaluation. The first derivatives
    and the cross derivative are imposed by the finite differences. The resulting
    function is C1.

    :param object x: An array-like object containing real values.
    :param object y: An array-like object containing real values.
    :param object f: A 2D array-like object of sample values corresponding to the
      `x` and `y` array points.
    :param bint extrapolate: If True, the extrapolation of data is enabled outside
      the range of the data set. The default is False. A ValueError is raised if
      extrapolation is disabled and a point is requested outside the data set.
    :param object extrapolation_type: Sets the method of extrapolation. The options
      are: 'nearest' (default), 'linear', 'quadratic'.
    :param double extrapolation_range: The attribute can be set to limit the range
      beyond the data set bounds that extrapolation is permitted. The default range
      is set to infinity. Requesting data beyond the extrapolation range will result
      in a ValueError being raised.
    :param tolerate_single_value: If True, single-value arrays will be tolerated as
      inputs. If a single value is supplied, that value will be extrapolated over
      the entire real range. If False (default), supplying a single value will
      result in a ValueError being raised.

    .. code-block:: pycon

       >>> import numpy as np
       >>> from cherab.core.math import Interpolate2DCubic
       >>>
       >>> # implements x**2 + y
       >>> drange = np.linspace(-2.5, 2.5, 100)
       >>> values = np.zeros((100, 100))
       >>> for i in range(100):
       >>>     for j in range(100):
       >>>         values[i, j] = drange[i]**2 + drange[j]
       >>>
       >>> f2d = Interpolate2DCubic(drange, drange, values)
       >>>
       >>> f2d(0, 0)
       0.00063769
       >>> f2d(-2, 1)
       5.00022956
       >>> f2d(-2, 3)
       ValueError: The specified value (x=-2.0, y=3.0) is outside the range of the supplied
       data and/or extrapolation range: x bounds=(-2.5, 2.5), y bounds=(-2.5, 2.5)
    """

    def __init__(self, object x, object y, object f, bint extrapolate=False, double extrapolation_range=float('inf'),
                 str extrapolation_type='nearest', bint tolerate_single_value=False):

        cdef int i, j, i_narrowed, j_narrowed

        supported_extrapolations = ['nearest', 'linear', 'quadratic']

        extrapolation_type = extrapolation_type.lower()
        if extrapolation_type not in supported_extrapolations:
            raise ValueError("Unsupported extrapolation type: {}".format(extrapolation_type))

        super().__init__(x, y, f, extrapolate, extrapolation_type, extrapolation_range, tolerate_single_value)


    cdef object _build(self, ndarray x, ndarray y, ndarray f):

        cdef:
            ndarray temp
            ndarray wx, wy, wf
            double[:,::1] f_mv
            int nx, ny
            int i, j, i_mapped, j_mapped

        # if x array is single valued, expand x array and data along x axis to simplify interpolation
        if x.shape[0] == 1:

            # set x array to full real range
            self._x = array([-INFINITY, INFINITY], dtype=float64)

            # duplicate data to provide a pseudo range for interpolation coefficient calculation
            temp = f
            x = array([-1., +1.], dtype=float64)
            f = empty((2, f.shape[1]), dtype=float64)
            f[:, :] = temp

        # if y array is single valued, expand y array and data along y axis to simplify interpolation
        if y.shape[0] == 1:

            # set y array to full real range
            self._y = array([-INFINITY, INFINITY], dtype=float64)

            # duplicate data to provide a pseudo range for interpolation coefficient calculation
            temp = f
            y = array([-1., +1.], dtype=float64)
            f = empty((f.shape[0], 2), dtype=float64)
            f[:, :] = temp

        nx = x.shape[0]
        ny = y.shape[0]

        # initialise the spline coefficient cache arrays
        self._k = empty((nx - 1, ny - 1, 16), dtype=float64)
        self._k[:,:,:] = NAN

        self._available = empty((nx - 1, ny - 1), dtype=int8)
        self._available[:,:] = False

        # normalise coordinate arrays
        self._ox = x.min()
        self._oy = y.min()

        self._sx = 1 / (x.max() - x.min())
        self._sy = 1 / (y.max() - y.min())

        x = (x - self._ox) * self._sx
        y = (y - self._oy) * self._sy

        # normalise data array
        self._of = f.min()
        self._sf = f.max() - f.min()
        if self._sf == 0:
            # zero data range, all values the same, disable scaling
            self._sf = 1
        f = (f - self._of) * (1 / self._sf)

        # widen arrays for automatic handling of boundaries polynomials
        wx = concatenate(([x[0]], x, [x[-1]]))
        wy = concatenate(([y[0]], y, [y[-1]]))
        wf = empty((nx + 2, ny + 2), dtype=float64)

        # store memory views to widened data
        self._wx = wx
        self._wy = wy
        self._wf = wf

        # populate expanded f array by duplicating the edges of the array
        f_mv = f
        for i in range(nx + 2):
            for j in range(ny + 2):
                i_mapped = min(max(0, i - 1), nx - 1)
                j_mapped = min(max(0, j - 1), ny - 1)
                self._wf[i, j] = f_mv[i_mapped, j_mapped]

        # calculate and cache higher powers of the dimension array data
        self._wx2 = wx * wx
        self._wx3 = wx * wx * wx

        self._wy2 = wy * wy
        self._wy3 = wy * wy * wy

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef double _evaluate(self, double px, double py, int order_x, int order_y, int ix, int iy) except? -1e999:
        """
        Evaluate the interpolating function which is valid in the area given
        by 'ix' and 'iy' at any position ('px', 'py').

        :param double px, double py: coordinates
        :param int ix, int iy: indices of the area of interest
        :return: the interpolated value
        """

        cdef:
            double px2, py2, px3, py3
            double[:,:,::1] k

        k = self._k

        # If the concerned polynomial has not yet been calculated:
        if not self._available[ix, iy]:
            self._calc_polynomial(ix, iy)

        if order_x == 0:

            # f(x,y)
            if order_y == 0:

                px2 = px*px
                px3 = px2*px
                py2 = py*py
                py3 = py2*py
                return     (k[ix, iy,  0] + k[ix, iy,  1]*py + k[ix, iy,  2]*py2 + k[ix, iy,  3]*py3) + \
                       px *(k[ix, iy,  4] + k[ix, iy,  5]*py + k[ix, iy,  6]*py2 + k[ix, iy,  7]*py3) + \
                       px2*(k[ix, iy,  8] + k[ix, iy,  9]*py + k[ix, iy, 10]*py2 + k[ix, iy, 11]*py3) + \
                       px3*(k[ix, iy, 12] + k[ix, iy, 13]*py + k[ix, iy, 14]*py2 + k[ix, iy, 15]*py3)

            # df(x,y) / dy
            if order_y == 1:

                px2 = px*px
                px3 = px2*px
                py2 = py*py
                return     (k[ix, iy,  1] + 2 * k[ix, iy,  2]*py + 3 * k[ix, iy,  3]*py2) + \
                       px *(k[ix, iy,  5] + 2 * k[ix, iy,  6]*py + 3 * k[ix, iy,  7]*py2) + \
                       px2*(k[ix, iy,  9] + 2 * k[ix, iy, 10]*py + 3 * k[ix, iy, 11]*py2) + \
                       px3*(k[ix, iy, 13] + 2 * k[ix, iy, 14]*py + 3 * k[ix, iy, 15]*py2)

            # d2f(x,y) / dy2
            elif order_y == 2:

                px2 = px*px
                px3 = px2*px
                return     (2 * k[ix, iy,  2] + 6 * k[ix, iy,  3]*py) + \
                       px *(2 * k[ix, iy,  6] + 6 * k[ix, iy,  7]*py) + \
                       px2*(2 * k[ix, iy, 10] + 6 * k[ix, iy, 11]*py) + \
                       px3*(2 * k[ix, iy, 14] + 6 * k[ix, iy, 15]*py)

            # d3f(x,y) / dy3
            elif order_y == 3:

                px2 = px*px
                px3 = px2*px
                return     (6 * k[ix, iy,  3]) + \
                       px *(6 * k[ix, iy,  7]) + \
                       px2*(6 * k[ix, iy, 11]) + \
                       px3*(6 * k[ix, iy, 15])

        elif order_x == 1:

            # df(x,y) / dx
            if order_y == 0:

                px2 = px*px
                py2 = py*py
                py3 = py2*py
                return          (k[ix, iy,  4] + k[ix, iy,  5]*py + k[ix, iy,  6]*py2 + k[ix, iy,  7]*py3) + \
                       px  * 2 *(k[ix, iy,  8] + k[ix, iy,  9]*py + k[ix, iy, 10]*py2 + k[ix, iy, 11]*py3) + \
                       px2 * 3 *(k[ix, iy, 12] + k[ix, iy, 13]*py + k[ix, iy, 14]*py2 + k[ix, iy, 15]*py3)

            # d2f(x,y) / dxdy
            elif order_y == 1:

                px2 = px*px
                py2 = py*py
                return          (k[ix, iy,  5] + 2 * k[ix, iy,  6]*py + 3 * k[ix, iy,  7]*py2) + \
                       px  * 2 *(k[ix, iy,  9] + 2 * k[ix, iy, 10]*py + 3 * k[ix, iy, 11]*py2) + \
                       px2 * 3 *(k[ix, iy, 13] + 2 * k[ix, iy, 14]*py + 3 * k[ix, iy, 15]*py2)

            # d3f(x,y) / dxdy2
            elif order_y == 2:

                px2 = px*px
                return          (2 * k[ix, iy,  6] + 6 * k[ix, iy,  7]*py) + \
                       px  * 2 *(2 * k[ix, iy, 10] + 6 * k[ix, iy, 11]*py) + \
                       px2 * 3 *(2 * k[ix, iy, 14] + 6 * k[ix, iy, 15]*py)

            # d4f(x,y) / dxdy3
            elif order_y == 3:

                px2 = px*px
                return (6 * k[ix, iy,  7]) + px  * 2 *(6 * k[ix, iy, 11]) + px2 * 3 *(6 * k[ix, iy, 15])

        elif order_x == 2:

            # d2f(x,y) / dx2
            if order_y == 0:

                py2 = py*py
                py3 = py2*py
                return      2 * (k[ix, iy,  8] + k[ix, iy,  9]*py + k[ix, iy, 10]*py2 + k[ix, iy, 11]*py3) + \
                       px * 6 * (k[ix, iy, 12] + k[ix, iy, 13]*py + k[ix, iy, 14]*py2 + k[ix, iy, 15]*py3)

            # d3f(x,y) / dx2dy
            elif order_y == 1:

                py2 = py*py
                return      2 * (k[ix, iy,  9] + 2 * k[ix, iy, 10]*py + 3 * k[ix, iy, 11]*py2) + \
                       px * 6 * (k[ix, iy, 13] + 2 * k[ix, iy, 14]*py + 3 * k[ix, iy, 15]*py2)

            # d4f(x,y) / dx2dy2
            elif order_y == 2:

                return      2 * (2 * k[ix, iy, 10] + 6 * k[ix, iy, 11]*py) + \
                       px * 6 * (2 * k[ix, iy, 14] + 6 * k[ix, iy, 15]*py)

            # d5f(x,y) / dx2dy3
            elif order_y == 3:
                return 2 * 6 * k[ix, iy, 11] + px * 6 * 6 * k[ix, iy, 15]

        elif order_x == 3:

            # d3f(x,y) / dx3
            if order_y == 0:

                py2 = py*py
                py3 = py2*py
                return 6 * (k[ix, iy, 12] + k[ix, iy, 13]*py + k[ix, iy, 14]*py2 + k[ix, iy, 15]*py3)

            # d4f(x,y) / dx3dy
            elif order_y == 1:

                py2 = py*py
                return 6 * (k[ix, iy, 13] + 2 * k[ix, iy, 14]*py + 3 * k[ix, iy, 15]*py2)

            # d5f(x,y) / dx3dy2
            elif order_y == 2:
                return 6 * (2 * k[ix, iy, 14] + 6 * k[ix, iy, 15]*py)

            # d6f(x,y) / dx3dy3
            elif order_y == 3:
                return 6 * 6 * k[ix, iy, 15]

        # higher orders
        return 0.0

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    @cython.cdivision(True)
    cdef int _calc_polynomial(self, int ix, int iy) except -1:
        """
        Calculates and caches the polynomial coefficients for area given by
        'ix', 'iy'. Declares this area as already calculated.

        :param int ix, int iy: indices of the area of interest
        """

        cdef:
            int u, v, l, i, j
            double delta_x, delta_y
            double cv_buffer[16]
            double cm_buffer[16][16]
            double[::1] cv, coeffs
            double[:,::1] cm
            double s

        # create memory views to constraint vector and matrix buffers
        cv = cv_buffer
        cm = cm_buffer

        # Fill the constraints matrix
        l = 0
        for u in range(ix+1, ix+3):
            for v in range(iy+1, iy+3):

                delta_x = self._wx[u+1] - self._wx[u-1]
                delta_y = self._wy[v+1] - self._wy[v-1]

                # knot values
                cm[l, 0] = 1.
                cm[l, 1] = self._wy[v]
                cm[l, 2] = self._wy2[v]
                cm[l, 3] = self._wy3[v]
                cm[l, 4] = self._wx[u]
                cm[l, 5] = self._wx[u]*self._wy[v]
                cm[l, 6] = self._wx[u]*self._wy2[v]
                cm[l, 7] = self._wx[u]*self._wy3[v]
                cm[l, 8] = self._wx2[u]
                cm[l, 9] = self._wx2[u]*self._wy[v]
                cm[l, 10] = self._wx2[u]*self._wy2[v]
                cm[l, 11] = self._wx2[u]*self._wy3[v]
                cm[l, 12] = self._wx3[u]
                cm[l, 13] = self._wx3[u]*self._wy[v]
                cm[l, 14] = self._wx3[u]*self._wy2[v]
                cm[l, 15] = self._wx3[u]*self._wy3[v]
                cv[l] = self._wf[u, v]
                l += 1

                # derivative along x
                cm[l, 0] = 0.
                cm[l, 1] = 0.
                cm[l, 2] = 0.
                cm[l, 3] = 0.
                cm[l, 4] = 1.
                cm[l, 5] = self._wy[v]
                cm[l, 6] = self._wy2[v]
                cm[l, 7] = self._wy3[v]
                cm[l, 8] = 2.*self._wx[u]
                cm[l, 9] = 2.*self._wx[u]*self._wy[v]
                cm[l, 10] = 2.*self._wx[u]*self._wy2[v]
                cm[l, 11] = 2.*self._wx[u]*self._wy3[v]
                cm[l, 12] = 3.*self._wx2[u]
                cm[l, 13] = 3.*self._wx2[u]*self._wy[v]
                cm[l, 14] = 3.*self._wx2[u]*self._wy2[v]
                cm[l, 15] = 3.*self._wx2[u]*self._wy3[v]
                cv[l] = (self._wf[u+1, v] - self._wf[u-1, v])/delta_x
                l += 1

                # derivative along y
                cm[l, 0] = 0.
                cm[l, 1] = 1.
                cm[l, 2] = 2.*self._wy[v]
                cm[l, 3] = 3.*self._wy2[v]
                cm[l, 4] = 0.
                cm[l, 5] = self._wx[u]
                cm[l, 6] = 2.*self._wx[u]*self._wy[v]
                cm[l, 7] = 3.*self._wx[u]*self._wy2[v]
                cm[l, 8] = 0.
                cm[l, 9] = self._wx2[u]
                cm[l, 10] = 2.*self._wx2[u]*self._wy[v]
                cm[l, 11] = 3.*self._wx2[u]*self._wy2[v]
                cm[l, 12] = 0.
                cm[l, 13] = self._wx3[u]
                cm[l, 14] = 2.*self._wx3[u]*self._wy[v]
                cm[l, 15] = 3.*self._wx3[u]*self._wy2[v]
                cv[l] = (self._wf[u, v+1] - self._wf[u, v-1])/delta_y
                l += 1

                # cross derivative
                cm[l, 0] = 0.
                cm[l, 1] = 0.
                cm[l, 2] = 0.
                cm[l, 3] = 0.
                cm[l, 4] = 0.
                cm[l, 5] = 1.
                cm[l, 6] = 2.*self._wy[v]
                cm[l, 7] = 3.*self._wy2[v]
                cm[l, 8] = 0.
                cm[l, 9] = 2.*self._wx[u]
                cm[l, 10] = 4.*self._wx[u]*self._wy[v]
                cm[l, 11] = 6.*self._wx[u]*self._wy2[v]
                cm[l, 12] = 0.
                cm[l, 13] = 3.*self._wx2[u]
                cm[l, 14] = 6.*self._wx2[u]*self._wy[v]
                cm[l, 15] = 9.*self._wx2[u]*self._wy2[v]
                cv[l] = (self._wf[u+1, v+1] - self._wf[u+1, v-1] - self._wf[u-1, v+1] + self._wf[u-1, v-1])/(delta_x*delta_y)
                l += 1

        # Solve the linear system and fill the caching coefficients array
        coeffs = solve(cm, cv)
        self._k[ix, iy, :] = coeffs

        # Denormalisation
        for i in range(4):
            for j in range(4):
                s = self._sf * self._sx**i * self._sy**j / (factorial(j) * factorial(i))
                coeffs[4*i + j] = s * self._calc_polynomial_derivative(ix, iy, -self._sx * self._ox, -self._sy * self._oy, i, j)
        coeffs[0] = coeffs[0] + self._of

        # populate coefficients and set cell as calculated
        self._k[ix, iy, :] = coeffs
        self._available[ix, iy] = True

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef double _extrapol_linear(self, double px, double py, int order_x, int order_y, int ix, int iy, double rx, double ry, bint inside_x, bint inside_y) except? -1e999:
        """
        Extrapolate linearly the interpolation function valid on area given by
        'ix' and 'iy' to position ('px', 'py').

        :param double px, double py: coordinates
        :param int ix, int iy: indices of the area of interest
        :param double rx, ry: the nearest position from ('px', 'py') in the interpolation domain.
        :return: the extrapolated value
        """

        cdef double ex, ey, result

        # calculate extrapolation distances from end of array
        ex = px - rx
        ey = py - ry

        # f(x,y)
        if order_x == 0 and order_y == 0:

            result = self._evaluate(rx, ry, 0, 0, ix, iy)
            if not inside_x:
                result += ex * self._evaluate(rx, ry, 1, 0, ix, iy)
            if not inside_y:
                result += ey * self._evaluate(rx, ry, 0, 1, ix, iy)
            return result

        # handle each domain separately
        if inside_x and not inside_y:

            if order_x == 0:

                # df(x,y)/dy
                if order_y == 1:
                    return self._evaluate(rx, ry, 0, 1, ix, iy)

                # dNf(x,y)/dyN
                else:
                    return 0

            if order_x == 1:

                # df(x,y)/dx
                if order_y == 0:
                    return self._evaluate(rx, ry, 1, 0, ix, iy) + ey * self._evaluate(rx, ry, 1, 1, ix, iy)

                # d2f(x,y)/dxdy
                elif order_y == 1:
                    return self._evaluate(rx, ry, 1, 1, ix, iy)

            if order_x == 2:

                # d2f(x,y)/dx2
                if order_y == 0:
                    return self._evaluate(rx, ry, 2, 0, ix, iy) + ey * self._evaluate(rx, ry, 2, 1, ix, iy)

                # d3f(fx,y)/dx2dy
                elif order_y == 1:
                    return self._evaluate(rx, ry, 2, 1, ix, iy)

            if order_x == 3:

                # d3f(x,y)/dx3
                if order_y == 0:
                    return self._evaluate(rx, ry, 3, 0, ix, iy) + ey * self._evaluate(rx, ry, 3, 1, ix, iy)

                # d4f(x,y)/dx3dy
                elif order_y == 1:
                    return self._evaluate(rx, ry, 3, 1, ix, iy)

            # higher orders
            return 0.0

        elif not inside_x and inside_y:

            if order_x == 0:

                # df(x,y)/dy
                if order_y == 1:
                    return self._evaluate(rx, ry, 0, 1, ix, iy) + ex * self._evaluate(rx, ry, 1, 1, ix, iy)

                # d2f(x,y)/dy
                elif order_y == 2:
                    return self._evaluate(rx, ry, 0, 2, ix, iy) + ex * self._evaluate(rx, ry, 1, 2, ix, iy)

                # d3f(x,y)/dy
                elif order_y == 3:
                    return self._evaluate(rx, ry, 0, 3, ix, iy) + ex * self._evaluate(rx, ry, 1, 3, ix, iy)

            elif order_x == 1:

                # df(x,y)/dx
                if order_y == 0:
                    return self._evaluate(rx, ry, 1, 0, ix, iy)

                # d2f(x,y)/dxdy
                elif order_y == 1:
                    return self._evaluate(rx, ry, 1, 1, ix, iy)

                # d3f(x,y)/dxdy2
                elif order_y == 2:
                    return self._evaluate(rx, ry, 1, 2, ix, iy)

                # d4f(x,y)/dxdy3
                elif order_y == 2:
                    return self._evaluate(rx, ry, 1, 3, ix, iy)

            # higher orders
            return 0.0

        elif not inside_x and not inside_y:

            if order_x == 0:

                # df(x,y)/dy
                if order_y == 1:
                    return self._evaluate(rx, ry, 0, 1, ix, iy)

            if order_x == 1:

                # df(x,y)/dx
                if order_y == 0:
                    return self._evaluate(rx, ry, 1, 0, ix, iy)

            # higher orders
            return 0.0

        raise RuntimeError('Extrapolation routine called for point in the interpolation domain.')

    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef double _extrapol_quadratic(self, double px, double py, int order_x, int order_y, int ix, int iy, double rx, double ry, bint inside_x, bint inside_y) except? -1e999:
        """
        Extrapolate quadratically the interpolation function valid on area given by
        'ix' and 'iy' to position ('px', 'py').

        :param double px, double py: coordinates
        :param int ix, int iy: indices of the area of interest
        :param double rx, ry: the nearest position from ('px', 'py') in the interpolation domain.
        :return: the extrapolated value
        """

        cdef double ex, ey, result

        # calculate extrapolation distances from end of array
        ex = px - rx
        ey = py - ry

        # f(x,y)
        if order_x == 0 and order_y == 0:

            result = self._evaluate(rx, ry, 0, 0, ix, iy)
            if not inside_x:
                result += ex * self._evaluate(rx, ry, 1, 0, ix, iy)
                result += 0.5 * ex*ex * self._evaluate(rx, ry, 2, 0, ix, iy)

            if not inside_y:
                result += ey * self._evaluate(rx, ry, 0, 1, ix, iy)
                result += 0.5 * ey*ey * self._evaluate(rx, ry, 0, 2, ix, iy)

            if not inside_x and not inside_y:
                result += ex*ey * self._evaluate(rx, ry, 1, 1, ix, iy)

            return result

        # handle each domain separately
        if inside_x and not inside_y:

            if order_x == 0:

                # df(x,y)/dy
                if order_y == 1:
                    return self._evaluate(rx, ry, 0, 1, ix, iy) + ey * self._evaluate(rx, ry, 0, 2, ix, iy)

                # d2f(x,y)/dy2
                elif order_y == 2:
                    return self._evaluate(rx, ry, 0, 2, ix, iy)

            elif order_x == 1:

                # df(x,y)/dx
                if order_y == 0:
                    return self._evaluate(rx, ry, 1, 0, ix, iy) \
                           + ey * self._evaluate(rx, ry, 1, 1, ix, iy) \
                           + 0.5 * ey*ey * self._evaluate(rx, ry, 1, 2, ix, iy)

                # d2f(x,y)/dxdy
                elif order_y == 1:
                    return self._evaluate(rx, ry, 1, 1, ix, iy) + ey * self._evaluate(rx, ry, 1, 2, ix, iy)

                # d3f(x,y)/dxdy2
                elif order_y == 2:
                    return self._evaluate(rx, ry, 1, 2, ix, iy)

            elif order_x == 2:

                # d2f(x,y)/dx2
                if order_y == 0:
                    return self._evaluate(rx, ry, 2, 0, ix, iy) \
                           + ey * self._evaluate(rx, ry, 2, 1, ix, iy) \
                           + 0.5 * ey*ey * self._evaluate(rx, ry, 2, 2, ix, iy)

                # d3f(x,y)/dx2dy
                elif order_y == 1:
                    return self._evaluate(rx, ry, 2, 1, ix, iy) + ey * self._evaluate(rx, ry, 2, 2, ix, iy)

                # d4f(x,y)/dx2dy2
                elif order_y == 2:
                    return self._evaluate(rx, ry, 2, 2, ix, iy)

            elif order_x == 3:

                # d3f(x,y)/dx3
                if order_y == 0:
                    return self._evaluate(rx, ry, 3, 0, ix, iy) \
                           + ey * self._evaluate(rx, ry, 3, 1, ix, iy) \
                           + 0.5 * ey*ey * self._evaluate(rx, ry, 3, 2, ix, iy)

                # d4f(x,y)/dx3dy
                elif order_y == 1:
                    return self._evaluate(rx, ry, 3, 1, ix, iy) + ey * self._evaluate(rx, ry, 3, 2, ix, iy)

                # d5f(x,y)/dx3dy2
                elif order_y == 2:
                    return self._evaluate(rx, ry, 3, 2, ix, iy)

            # higher orders
            return 0.0

        elif not inside_x and inside_y:

            if order_x == 0:

                # df(x,y)/dy
                if order_y == 1:
                    return self._evaluate(rx, ry, 0, 1, ix, iy) \
                           + ex * self._evaluate(rx, ry, 1, 1, ix, iy) \
                           + 0.5 * ex*ex * self._evaluate(rx, ry, 2, 1, ix, iy)

                # d2f(x,y)/dy2
                elif order_y == 2:
                    return self._evaluate(rx, ry, 0, 2, ix, iy) \
                           + ex * self._evaluate(rx, ry, 1, 2, ix, iy) \
                           + 0.5 * ex*ex * self._evaluate(rx, ry, 2, 2, ix, iy)


                # d3f(x,y)/dy3
                elif order_y == 3:
                    return self._evaluate(rx, ry, 0, 3, ix, iy) \
                           + ex * self._evaluate(rx, ry, 1, 3, ix, iy) \
                           + 0.5 * ex*ex * self._evaluate(rx, ry, 2, 3, ix, iy)

            elif order_x == 1:

                # df(x,y)/dx
                if order_y == 0:
                    return self._evaluate(rx, ry, 1, 0, ix, iy) + ex * self._evaluate(rx, ry, 2, 0, ix, iy)

                # d2f(x,y)/dxdy
                elif order_y == 1:
                    return self._evaluate(rx, ry, 1, 1, ix, iy) + ex * self._evaluate(rx, ry, 2, 1, ix, iy)

                # d3f(x,y)/dxdy2
                elif order_y == 2:
                    return self._evaluate(rx, ry, 1, 2, ix, iy) + ex * self._evaluate(rx, ry, 2, 2, ix, iy)

                # d4f(x,y)/dxdy3
                elif order_y == 3:
                    return self._evaluate(rx, ry, 1, 3, ix, iy) + ex * self._evaluate(rx, ry, 2, 3, ix, iy)

            elif order_x == 2:

                # d2f(x,y)/dx2
                if order_y == 0:
                    return self._evaluate(rx, ry, 2, 0, ix, iy)

                # d3f(x,y)/dx2dy
                elif order_y == 1:
                    return self._evaluate(rx, ry, 2, 1, ix, iy)

                # d4f(x,y)/dx2dy2
                elif order_y == 2:
                    return self._evaluate(rx, ry, 2, 2, ix, iy)

                # d5f(x,y)/dx2dy3
                elif order_y == 3:
                    return self._evaluate(rx, ry, 2, 3, ix, iy)

            # higher orders
            return 0.0

        elif not inside_x and not inside_y:

            if order_x == 0:

                # df(x,y)/dy
                if order_y == 1:
                    return self._evaluate(rx, ry, 0, 1, ix, iy) \
                           + ey * self._evaluate(rx, ry, 0, 2, ix, iy) \
                           + ex * self._evaluate(rx, ry, 1, 1, ix, iy)

                # d2f(x,y)/dy2
                elif order_y == 2:
                    return self._evaluate(rx, ry, 0, 2, ix, iy)

            elif order_x == 1:

                # df(x,y)/dx
                if order_y == 0:
                    return self._evaluate(rx, ry, 1, 0, ix, iy) \
                           + ex * self._evaluate(rx, ry, 2, 0, ix, iy) \
                           + ey * self._evaluate(rx, ry, 1, 1, ix, iy)

                # d2f(x,y)/dxdy
                elif order_y == 1:
                    return self._evaluate(rx, ry, 1, 1, ix, iy)

            elif order_x == 2:

                # d2f(x,y)/dx2
                if order_y == 0:
                    return self._evaluate(rx, ry, 2, 0, ix, iy)

            # higher orders
            return 0.0

        raise RuntimeError('Extrapolation routine called for point in the interpolation domain.')

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef double _calc_polynomial_derivative(self, int ix, int iy, double px, double py, int order_x, int order_y):
        """
        Evaluate the derivatives of the polynomial valid in the area given by
        'ix', 'iy' at position ('px', 'py'). The order of
        derivative along each axis is given by 'der_x', 'der_y'.

        :param int ix, int iy: indices of the area of interest
        :param double px, double py: coordinates
        :param int der_x, int order_y: orders of derivative along each axis
        :return: value evaluated from the derivated polynomial
        """

        cdef:
            double[::1] ax, ay
            double[:,:,::1] k

        k = self._k

        ax = derivatives_array(px, order_x)
        ay = derivatives_array(py, order_y)

        return ax[0]*(ay[0]*k[ix, iy,  0] + ay[1]*k[ix, iy,  1] + ay[2]*k[ix, iy,  2] + ay[3]*k[ix, iy,  3]) + \
               ax[1]*(ay[0]*k[ix, iy,  4] + ay[1]*k[ix, iy,  5] + ay[2]*k[ix, iy,  6] + ay[3]*k[ix, iy,  7]) + \
               ax[2]*(ay[0]*k[ix, iy,  8] + ay[1]*k[ix, iy,  9] + ay[2]*k[ix, iy, 10] + ay[3]*k[ix, iy, 11]) + \
               ax[3]*(ay[0]*k[ix, iy, 12] + ay[1]*k[ix, iy, 13] + ay[2]*k[ix, iy, 14] + ay[3]*k[ix, iy, 15])



# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

from raysect.core.math.function.float cimport Blend1D, Blend2D, Blend3D
from cherab.core.math.samplers cimport *
from cherab.core.math.function cimport *
from cherab.core.math.interpolators cimport *
from cherab.core.math.caching cimport *
from cherab.core.math.clamp cimport *
from cherab.core.math.mappers cimport *
from cherab.core.math.mask cimport *
from cherab.core.math.slice cimport *
from cherab.core.math.transform cimport *



# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

"""
Demonstration of RayTransferCylinder in axisymmetrical case and applied voxel map
---------------------------------------------------------------------------------

This file will demonstrate how to:

 * calculate ray transfer matrix (geometry matrix) for axisymmetrical cylindrical emitter defined
   on a regular grid
 * map multiple grid cells to a single light source by applying a voxel map

"""
import os
import numpy as np
from matplotlib import pyplot as plt

from raysect.primitive import Cylinder, Subtract
from raysect.optical import World, translate, rotate
from raysect.optical.observer import PinholeCamera, FullFrameSampler2D

from cherab.tools.raytransfer import RayTransferPipeline2D, RayTransferCylinder
from cherab.tools.raytransfer import RoughNickel

world = World()

# creating the scene
cylinder_inner = Cylinder(radius=80., height=140.)
cylinder_outer = Cylinder(radius=220., height=140.)
wall = Subtract(cylinder_outer, cylinder_inner, material=RoughNickel(0.1), parent=world,
                transform=translate(0, 0, -70.))

# creating ray transfer cylinder with 200 (m) outer radius, 100 (m) inner radius, 140 (m) height
# for axisymmetric cylindrical emission profile defined on a 100 x 100 (R, Z) gird
rtc = RayTransferCylinder(200., 100., 100, 100, radius_inner=100.)
# n_polar=0 by default (axisymmetric case)
rtc.parent = world
rtc.transform = translate(0, 0, -50.)

# unlike the demo with a mask, here we not only cut out a circle but also
# create 50 ring-shaped light sources using the voxel map
rad_circle = 50.
xsqr = np.linspace(-49.5, 49.5, 100) ** 2
rad = np.sqrt(xsqr[:, None] + xsqr[None, :])
voxel_map = np.zeros((100, 100), dtype=np.int32)
voxel_map[rad > 50.] = -1  # removing the area outside the circle
for i in range(50):
    voxel_map[(rad < i + 1.) * (rad > i)] = i  # mapping multiple grid cells to a single light source
rtc.voxel_map = voxel_map[:, None, :]  # making 3D voxel map from 2D (RZ-plane) voxel map and applying it
# now we have only 50 light sources

# creating ray transfer pipeline
pipeline = RayTransferPipeline2D()

# setting up the camera
camera = PinholeCamera((256, 256), pipelines=[pipeline], frame_sampler=FullFrameSampler2D(),
                       transform=translate(219., 0, 0) * rotate(90., 0., -90.), parent=world)
camera.fov = 90
camera.pixel_samples = 500
camera.min_wavelength = 500.
camera.max_wavelength = camera.min_wavelength + 1.
camera.spectral_bins = rtc.bins

# starting ray tracing
camera.observe()

# uncomment this to save ray transfer matrix to file
# np.save('ray_transfer_map.npy', pipeline.matrix)

# let's collapse the ray transfer matrix with some emission profiles

# obtaining 30 images for 30 emission profiles
images = []
vmax = 0
rad_inside = np.linspace(0.5, 49.5, 50)  # we have only 50 light sources now
shifts = np.linspace(0, 2. / 3., 30, endpoint=False)
for shift in shifts:
    profile = ((1. + np.cos(3. * np.pi * (rad_inside / rad_circle - shift))) *
               np.exp(-rad_inside / rad_circle))
    image = np.dot(pipeline.matrix, profile)
    vmax = max(vmax, image.max())
    images.append(image)

# creating a folder for images
if not os.path.exists('images'):
    os.makedirs('images')

# now let's see all 30 ray-traced images obtained with only a single run of ray-tracing
fig = plt.figure(figsize=(4., 4.))
ax = fig.add_subplot(111)
fig.subplots_adjust(left=0.05, right=0.95, bottom=0.05, top=0.95)
for i in range(30):
    ax.cla()
    ax.imshow(images[i].T, cmap='gray', vmax=0.75 * vmax)
    ax.tick_params(labelbottom=False, labelleft=False, bottom=False, left=False)
    fig.savefig('images/ray_transfer_map_demo_%02d.png' % i, dpi=180)
    plt.pause(0.1)

# creating gif animation with ImageMagick
os.system("convert -delay 10 -loop 0 images/ray_transfer_map_demo_*.png ray_transfer_map_demo.gif")
# compare with ray_transfer_mask_demo.gif




# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

from raysect.core.ray import Ray as CoreRay
from raysect.optical.material.material import NullMaterial


def find_wall_intersection(world, centre_point, sightline_vec, delta=1E-3):

    while True:

        # Find the next intersection point of the ray with the world
        intersection = world.hit(CoreRay(centre_point, sightline_vec))

        if intersection is None:
            raise ValueError('No intersection with solid material found.')

        elif isinstance(intersection.primitive.material, NullMaterial):
            centre_point += sightline_vec * delta
            continue

        else:
            hit_point = intersection.hit_point.transform(intersection.primitive_to_world)
            return hit_point, intersection.primitive



# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

import sys

import numpy as np
from scipy.constants import electron_mass, atomic_mass
from matplotlib.pyplot import ion, ioff, plot, show

from raysect.optical.library import *
from raysect.primitive import Sphere, Box, Intersect
from raysect.optical import Ray, d65_white, World, Point3D, Vector3D, translate, rotate
from raysect.optical.observer import PinholeCamera
from raysect.optical.material.emitter import Checkerboard

from cherab.core import Plasma, Beam, Species, Maxwellian
from cherab.core.atomic import elements, Line
from cherab.openadas import OpenADAS
from cherab.core.model import SingleRayAttenuator, BeamCXLine
from cherab.tools.plasmas import GaussianVolume

integration_step = 0.1

# setup scenegraph
world = World()

# create atomic data source
adas = OpenADAS(permit_extrapolation=True)

# PLASMA ----------------------------------------------------------------------
plasma = Plasma(parent=world)

# define basic distributions
ion_density = 9e19
sigma = 0.25

d_density = GaussianVolume(0.94 * ion_density, sigma)
he2_density = GaussianVolume(0.04 * ion_density, sigma)
c6_density = GaussianVolume(0.01 * ion_density, sigma)
ne10_density = GaussianVolume(0.01 * ion_density, sigma)
e_density = GaussianVolume((0.94 + 0.04*2 + 0.01*6 + 0.01*10) * ion_density, sigma)
temperature = 10 + GaussianVolume(240, sigma)
# temperature = 1000 + GaussianVolume(4000, sigma)
bulk_velocity = Vector3D(200e3, 0, 0)

d_distribution = Maxwellian(d_density, temperature, bulk_velocity, elements.deuterium.atomic_weight * atomic_mass)
he2_distribution = Maxwellian(he2_density, temperature, bulk_velocity, elements.helium.atomic_weight * atomic_mass)
c6_distribution = Maxwellian(c6_density, temperature, bulk_velocity, elements.carbon.atomic_weight * atomic_mass)
ne10_distribution = Maxwellian(ne10_density, temperature, bulk_velocity, elements.neon.atomic_weight * atomic_mass)
e_distribution = Maxwellian(e_density, temperature, bulk_velocity, electron_mass)

d_species = Species(elements.deuterium, 1, d_distribution)
he2_species = Species(elements.helium, 2, he2_distribution)
c6_species = Species(elements.carbon, 6, c6_distribution)
ne10_species = Species(elements.neon, 10, ne10_distribution)

# define species
plasma.b_field = Vector3D(1.0, 1.0, 1.0)
plasma.electron_distribution = e_distribution
plasma.composition = [d_species, he2_species, c6_species, ne10_species]

print("plasma.z_effective(): ", plasma.z_effective(0, 0, 0))


from cherab.core.model import ExcitationLine, RecombinationLine

plasma.geometry = Sphere(sigma * 5.0)
plasma.geometry_transform = None
plasma.integrator.step = integration_step
plasma.integrator.min_samples = 5
plasma.atomic_data = adas

# Setup elements.deuterium lines
d_alpha = Line(elements.deuterium, 0, (3, 2))
d_beta = Line(elements.deuterium, 0, (4, 2))
d_gamma = Line(elements.deuterium, 0, (5, 2))
d_delta = Line(elements.deuterium, 0, (6, 2))
d_epsilon = Line(elements.deuterium, 0, (7, 2))

plasma.models = [
    # Bremsstrahlung(),
    RecombinationLine(d_alpha),
    RecombinationLine(d_beta),
    RecombinationLine(d_gamma),
    RecombinationLine(d_delta),
    RecombinationLine(d_epsilon)
]


# BEAM ------------------------------------------------------------------------
beam = Beam(parent=world, transform=translate(1.0, 0.0, 0) * rotate(90, 0, 0))
beam.plasma = plasma
beam.atomic_data = adas
beam.energy = 60000
beam.power = 1e4
beam.element = elements.deuterium
beam.sigma = 0.025
beam.divergence_x = 0.5
beam.divergence_y = 0.5
beam.length = 3.0
beam.attenuator = SingleRayAttenuator(clamp_to_zero=True)
beam.models = [
    BeamCXLine(Line(elements.helium, 1, (4, 3))),
    BeamCXLine(Line(elements.helium, 1, (6, 4))),
    BeamCXLine(Line(elements.carbon, 5, (8, 7))),
    BeamCXLine(Line(elements.carbon, 5, (9, 8))),
    BeamCXLine(Line(elements.carbon, 5, (10, 8))),
    BeamCXLine(Line(elements.neon, 9, (11, 10))),
    BeamCXLine(Line(elements.neon, 9, (12, 11))),
]
beam.integrator.step = integration_step
beam.integrator.min_samples = 5

beam = Beam(parent=world, transform=translate(1.0, 0.0, 0) * rotate(90, 0, 0))
beam.plasma = plasma
beam.atomic_data = adas
beam.energy = 60000 / 2
beam.power = 1e4
beam.element = elements.deuterium
beam.sigma = 0.025
beam.divergence_x = 0.5
beam.divergence_y = 0.5
beam.length = 3.0
beam.attenuator = SingleRayAttenuator(clamp_to_zero=True)
beam.models = [
    BeamCXLine(Line(elements.helium, 1, (4, 3))),
    BeamCXLine(Line(elements.helium, 1, (6, 4))),
    BeamCXLine(Line(elements.carbon, 5, (8, 7))),
    BeamCXLine(Line(elements.carbon, 5, (9, 8))),
    BeamCXLine(Line(elements.carbon, 5, (10, 8))),
    BeamCXLine(Line(elements.neon, 9, (11, 10))),
    BeamCXLine(Line(elements.neon, 9, (12, 11))),
]
beam.integrator.step = integration_step
beam.integrator.min_samples = 5

beam = Beam(parent=world, transform=translate(1.0, 0.0, 0) * rotate(90, 0, 0))
beam.plasma = plasma
beam.atomic_data = adas
beam.energy = 60000 / 3
beam.power = 1e4
beam.element = elements.deuterium
beam.sigma = 0.025
beam.divergence_x = 0.5
beam.divergence_y = 0.5
beam.length = 3.0
beam.attenuator = SingleRayAttenuator(clamp_to_zero=True)
beam.models = [
    BeamCXLine(Line(elements.helium, 1, (4, 3))),
    BeamCXLine(Line(elements.helium, 1, (6, 4))),
    BeamCXLine(Line(elements.carbon, 5, (8, 7))),
    BeamCXLine(Line(elements.carbon, 5, (9, 8))),
    BeamCXLine(Line(elements.carbon, 5, (10, 8))),
    BeamCXLine(Line(elements.neon, 9, (11, 10))),
    BeamCXLine(Line(elements.neon, 9, (12, 11))),
]
beam.integrator.step = integration_step
beam.integrator.min_samples = 5

# LENS ------------------------------------------------------------------------

# s1 = Sphere(1.0, transform=translate(0, 0, 1.0-0.01))
# s2 = Sphere(0.5, transform=translate(0, 0, -0.5+0.01))
# Intersect(s1, s2, world, translate(0, 0, -1.6)*rotate(-30,30,0), schott("N-BK7"))

# BACKGROUND ------------------------------------------------------------------

# Box(Point3D(-50, -50, 50), Point3D(50, 50, 50.1), world, material=Checkerboard(4, d65_white, d65_white, 0.001, 0.002))
# Box(Point(-100, -100, -100), Point(100, 100, 100), world, material=UniformSurfaceEmitter(d65_white, 0.001))

# OBSERVER --------------------------------------------------------------------

#import cProfile
#
# def profile_test(n=25000):
#     r = Ray(origin=Point(0.0, 0, 0), min_wavelength=526, max_wavelength=532, num_samples=100)
#     for i in range(n):
#         r.trace(world)
#
# cProfile.run("profile_test()", sort="tottime")

ion()

r = Ray(origin=Point3D(0.5, 0, -2.5), min_wavelength=440, max_wavelength=740, bins=800)
s = r.trace(world)
plot(s.wavelengths, s.samples)

r = Ray(origin=Point3D(0.5, 0, -2.5), min_wavelength=440, max_wavelength=740, bins=3200)
s = r.trace(world)
plot(s.wavelengths, s.samples)
show()

camera = PinholeCamera((128, 128), parent=world, transform=translate(0, 0, -2.5))
camera.spectral_rays = 1
camera.spectral_bins = 15
camera.pixel_samples = 10

ion()
camera.observe()

ioff()
camera.pipelines[0].display()



# Copyright 2016-2023 Euratom
# Copyright 2016-2023 United Kingdom Atomic Energy Authority
# Copyright 2016-2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

import unittest

import numpy as np

from raysect.core import Point3D, Vector3D, translate
from raysect.optical import World, Spectrum, Ray

from cherab.core import Beam
from cherab.core.atomic import Line, AtomicData, BeamCXPEC, BeamStoppingRate
from cherab.core.atomic import deuterium
from cherab.tools.plasmas.slab import build_constant_slab_plasma
from cherab.core.model import SingleRayAttenuator, BeamCXLine, GaussianLine, ZeemanTriplet


class ConstantBeamCXPEC(BeamCXPEC):
    """
    Constant beam CX PEC for test purpose.
    """

    def __init__(self, donor_metastable, value):
        super().__init__(donor_metastable)
        self.value = value

    def evaluate(self, energy, temperature, density, z_effective, b_field):

        return self.value


class ConstantBeamStoppingRate(BeamStoppingRate):
    """
    Constant beam CX PEC for test purpose.
    """

    def __init__(self, donor_metastable, value):
        self.donor_metastable = donor_metastable
        self.value = value

    def evaluate(self, energy, density, temperature):

        return self.value


class MockAtomicData(AtomicData):
    """Fake atomic data for test purpose."""

    def beam_cx_pec(self, donor_ion, receiver_ion, receiver_charge, transition):

        return [ConstantBeamCXPEC(1, 3.4e-34)]

    def beam_stopping_rate(self, beam_ion, plasma_ion, charge):

        return ConstantBeamStoppingRate(1, 0)

    def wavelength(self, ion, charge, transition):

        return 656.104


class TestBeamCXLine(unittest.TestCase):

    def setUp(self):

        self.world = World()

        self.atomic_data = MockAtomicData()

        plasma_species = [(deuterium, 1, 1.e19, 200., Vector3D(0, 0, 0))]
        plasma = build_constant_slab_plasma(length=1, width=1, height=1,
                                            electron_density=1e19,
                                            electron_temperature=200.,
                                            plasma_species=plasma_species,
                                            b_field=Vector3D(0, 10., 0))
        plasma.atomic_data = self.atomic_data
        plasma.parent = self.world

        beam = Beam(transform=translate(0.5, 0, 0))
        beam.atomic_data = self.atomic_data
        beam.plasma = plasma
        beam.attenuator = SingleRayAttenuator(clamp_to_zero=True)
        beam.energy = 50000
        beam.power = 1e6
        beam.temperature = 10
        beam.element = deuterium
        beam.parent = self.world

        self.plasma = plasma
        self.beam = beam

    def test_default_lineshape(self):
        # setting up the model
        line = Line(deuterium, 0, (3, 2))  # D-alpha line
        self.beam.models = [BeamCXLine(line)]

        # observing
        origin = Point3D(1.5, 0, 0)
        direction = Vector3D(-1, 0, 0)
        ray = Ray(origin=origin, direction=direction,
                  min_wavelength=655.1, max_wavelength=657.1, bins=512)
        cx_spectrum = ray.trace(self.world)

        # validating
        dx = self.beam.integrator.step
        rate = self.atomic_data.beam_cx_pec(deuterium, deuterium, 1, (3, 2))[0].value
        ni = self.plasma.ion_density(0.5, 0, 0)  # constant slab
        nd_beam = 0  # beam density
        for i in range(-int(0.5 / dx), int(0.5 / dx)):
            x = dx * (i + 0.5)
            nd_beam += self.beam.density(x, 0, 0)  # in internal beam coordinates
        radiance = 0.25 * rate * ni * nd_beam * dx / np.pi

        target_species = self.plasma.composition.get(line.element, line.charge + 1)
        wavelength = self.atomic_data.wavelength(line.element, line.charge, line.transition)
        gaussian_line = GaussianLine(line, wavelength, target_species, self.plasma, self.atomic_data)
        spectrum = Spectrum(ray.min_wavelength, ray.max_wavelength, ray.bins)
        spectrum = gaussian_line.add_line(radiance, Point3D(0.5, 0, 0), direction, spectrum)

        for i in range(ray.bins):
            self.assertAlmostEqual(cx_spectrum.samples[i], spectrum.samples[i], delta=1e-8,
                                   msg='BeamCXLine model gives a wrong value at {} nm.'.format(spectrum.wavelengths[i]))

    def test_custom_lineshape(self):
        # setting up the model
        line = Line(deuterium, 0, (3, 2))  # D-alpha line
        self.beam.models = [BeamCXLine(line, lineshape=ZeemanTriplet)]

        # observing
        origin = Point3D(1.5, 0, 0)
        direction = Vector3D(-1, 0, 0)
        ray = Ray(origin=origin, direction=direction,
                  min_wavelength=655.1, max_wavelength=657.1, bins=512)
        cx_spectrum = ray.trace(self.world)

        # validating
        dx = self.beam.integrator.step
        rate = self.atomic_data.beam_cx_pec(deuterium, deuterium, 1, (3, 2))[0].value
        ni = self.plasma.ion_density(0.5, 0, 0)  # constant slab
        nd_beam = 0  # beam density
        for i in range(-int(0.5 / dx), int(0.5 / dx)):
            x = dx * (i + 0.5)
            nd_beam += self.beam.density(x, 0, 0)  # in internal beam coordinates
        radiance = 0.25 * rate * ni * nd_beam * dx / np.pi

        target_species = self.plasma.composition.get(line.element, line.charge + 1)
        wavelength = self.atomic_data.wavelength(line.element, line.charge, line.transition)
        zeeman_line = ZeemanTriplet(line, wavelength, target_species, self.plasma, self.atomic_data)
        spectrum = Spectrum(ray.min_wavelength, ray.max_wavelength, ray.bins)
        spectrum = zeeman_line.add_line(radiance, Point3D(0.5, 0, 0), direction, spectrum)

        for i in range(ray.bins):
            self.assertAlmostEqual(cx_spectrum.samples[i], spectrum.samples[i], delta=1e-8,
                                   msg='BeamCXLine model gives a wrong value at {} nm.'.format(spectrum.wavelengths[i]))


if __name__ == '__main__':
    unittest.main()



# cython: language_level=3

# Copyright 2016-2023 Euratom
# Copyright 2016-2023 United Kingdom Atomic Energy Authority
# Copyright 2016-2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

from libc.math cimport fabs, sqrt
from raysect.optical cimport Spectrum, Point3D, Vector3D

from cherab.core.plasma cimport Plasma
from cherab.core.beam cimport Beam
from cherab.core.atomic cimport AtomicData
from cherab.core.atomic cimport Line
from cherab.core.math.function cimport autowrap_function1d, autowrap_function2d
from cherab.core.utility.constants cimport ATOMIC_MASS, ELEMENTARY_CHARGE
from cherab.core.model.lineshape.gaussian cimport add_gaussian_line
from cherab.core.model.lineshape.doppler cimport thermal_broadening, doppler_shift

cimport cython


cdef double RECIP_ATOMIC_MASS = 1 / ATOMIC_MASS


cdef double evamu_to_ms(double x):
    return sqrt(2 * x * ELEMENTARY_CHARGE * RECIP_ATOMIC_MASS)


DEF STARK_SPLITTING_FACTOR = 2.77e-8


cdef class BeamEmissionMultiplet(BeamLineShapeModel):
    """
    Produces Beam Emission Multiplet line shape, also known as the Motional Stark Effect spectrum.
    """

    def __init__(self, Line line, double wavelength, Beam beam, AtomicData atomic_data,
                 object sigma_to_pi, object sigma1_to_sigma0, object pi2_to_pi3, object pi4_to_pi3):

        super().__init__(line, wavelength, beam, atomic_data)

        self._sigma_to_pi = autowrap_function2d(sigma_to_pi)
        self._sigma1_to_sigma0 = autowrap_function1d(sigma1_to_sigma0)
        self._pi2_to_pi3 = autowrap_function1d(pi2_to_pi3)
        self._pi4_to_pi3 = autowrap_function1d(pi4_to_pi3)

    @cython.cdivision(True)
    cpdef Spectrum add_line(self, double radiance, Point3D beam_point, Point3D plasma_point,
                            Vector3D beam_direction, Vector3D observation_direction, Spectrum spectrum):

        cdef double x, y, z
        cdef Plasma plasma
        cdef double te, ne, beam_energy, sigma, stark_split, beam_ion_mass, beam_temperature
        cdef double natural_wavelength, central_wavelength
        cdef double sigma_to_pi, d, intensity_sig, intensity_pi, e_field
        cdef double s1_to_s0, intensity_s0, intensity_s1
        cdef double pi2_to_pi3, pi4_to_pi3, intensity_pi2, intensity_pi3, intensity_pi4
        cdef Vector3D b_field, beam_velocity

        # extract for more compact code
        x = plasma_point.x
        y = plasma_point.y
        z = plasma_point.z

        plasma = self.beam.get_plasma()

        te = plasma.get_electron_distribution().effective_temperature(x, y, z)
        if te <= 0.0:
            return spectrum

        ne = plasma.get_electron_distribution().density(x, y, z)
        if ne <= 0.0:
            return spectrum

        beam_energy = self.beam.get_energy()

        # calculate Stark splitting
        b_field = plasma.get_b_field().evaluate(x, y, z)
        beam_velocity = beam_direction.normalise().mul(evamu_to_ms(beam_energy))
        e_field = beam_velocity.cross(b_field).get_length()
        stark_split = fabs(STARK_SPLITTING_FACTOR * e_field)  # TODO - calculate splitting factor? Reject other lines?

        # calculate emission line central wavelength, doppler shifted along observation direction
        natural_wavelength = self.wavelength
        central_wavelength = doppler_shift(natural_wavelength, observation_direction, beam_velocity)

        # calculate doppler broadening
        beam_ion_mass = self.beam.get_element().atomic_weight
        beam_temperature = self.beam.get_temperature()
        sigma = thermal_broadening(self.wavelength, beam_temperature, beam_ion_mass)

        # calculate relative intensities of sigma and pi lines
        sigma_to_pi = self._sigma_to_pi.evaluate(ne, beam_energy)
        d = 1 / (1 + sigma_to_pi)
        intensity_sig = sigma_to_pi * d * radiance
        intensity_pi = 0.5 * d * radiance

        # add Sigma lines to output
        s1_to_s0 = self._sigma1_to_sigma0.evaluate(ne)
        intensity_s0 = 1 / (s1_to_s0 + 1)
        intensity_s1 = 0.5 * s1_to_s0 * intensity_s0

        spectrum = add_gaussian_line(intensity_sig * intensity_s0, central_wavelength, sigma, spectrum)
        spectrum = add_gaussian_line(intensity_sig * intensity_s1, central_wavelength + stark_split, sigma, spectrum)
        spectrum = add_gaussian_line(intensity_sig * intensity_s1, central_wavelength - stark_split, sigma, spectrum)

        # add Pi lines to output
        pi2_to_pi3 = self._pi2_to_pi3.evaluate(ne)
        pi4_to_pi3 = self._pi4_to_pi3.evaluate(ne)
        intensity_pi3 = 1 / (1 + pi2_to_pi3 + pi4_to_pi3)
        intensity_pi2 = pi2_to_pi3 * intensity_pi3
        intensity_pi4 = pi4_to_pi3 * intensity_pi3

        spectrum = add_gaussian_line(intensity_pi * intensity_pi2, central_wavelength + 2 * stark_split, sigma, spectrum)
        spectrum = add_gaussian_line(intensity_pi * intensity_pi2, central_wavelength - 2 * stark_split, sigma, spectrum)
        spectrum = add_gaussian_line(intensity_pi * intensity_pi3, central_wavelength + 3 * stark_split, sigma, spectrum)
        spectrum = add_gaussian_line(intensity_pi * intensity_pi3, central_wavelength - 3 * stark_split, sigma, spectrum)
        spectrum = add_gaussian_line(intensity_pi * intensity_pi4, central_wavelength + 4 * stark_split, sigma, spectrum)
        spectrum = add_gaussian_line(intensity_pi * intensity_pi4, central_wavelength - 4 * stark_split, sigma, spectrum)

        return spectrum



# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

cdef class PlasmaModel:
    """
    A plasma emission model.

    When attached to a plasma, a plasma emission model samples the plasma properties
    and atomic data it needs to calculate its emission. The emission is calculated
    for a particular point and viewing orientation in plasma space.

    A new emission model is implemented by inheriting from this class and specifying
    the emission() function.

    If it is necessary to cache data to speed up the emission
    calculation and there is a risk the cached data may be made stale by changes to the
    plasma, the _change() method must be implemented to reset the cache. The _change()
    function is automatically called when changes occur on the Plasma object.

    The plasma and atomic data provider attributes will be automatically populated
    when the PlasmaModel is attached to the Plasma object. In general these should
    not be set by the user directly.

    :ivar Plasma plasma: The plasma to which this emission model is attached.
    :ivar AtomicData atomic_data: The atomic data provider for this model.
    """

    def __init__(self, Plasma plasma=None, AtomicData atomic_data=None):

        self._plasma = plasma
        self._atomic_data = atomic_data

        # setup change notification is we have been given a plasma object
        if self._plasma:
            self._plasma.notifier.add(self._change)

    @property
    def plasma(self):
        return self._plasma

    @plasma.setter
    def plasma(self, value):

        # disconnect from previous plasma's notifications
        if self._plasma:
            self._plasma.notifier.remove(self._change)

        # attach to plasma to inform model of changes to plasma properties
        self._plasma = value
        self._plasma.notifier.add(self._change)

        # inform model source data has changed
        self._change()

    @property
    def atomic_data(self):
        return self._atomic_data

    @atomic_data.setter
    def atomic_data(self, value):

        self._atomic_data = value

        # inform model source data has changed
        self._change()

    cpdef Spectrum emission(self, Point3D point, Vector3D direction, Spectrum spectrum):
        """
        Calculate the emission for a point in the plasma in a specified direction.

        Models implementing this method must add their spectral response to the
        supplied spectrum object. The spectrum units are spectral radiance per
        meter (W/m^3/str/nm).
        
        If a model has a directional response, the model should pass through
        its own reference axis e.g. Thomsen scattering laser direction. 
                       
        :param point: Point in plasma space.
        :param direction: Direction in plasma space.
        :param spectrum: Spectrum to which emission should be added.
        :return: Updated Spectrum object.
        """

        raise NotImplementedError('Virtual method must be implemented in a sub-class.')

    def _change(self):
        """
        Called if the plasma properties or the atomic data source changes.

        If the model caches calculation data that would be invalidated if its
        source data changes then this method may be overridden to clear the
        cache.

        This method is triggered if the plasma notifies the model of a change
        or the atomic data source is changed.
        """

        pass




# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

"""
This script has been used to calculate the reference data for the 2D interpolators tests.
To get the reference data for the 2D linear interpolators, uncomment the linear
part (l.106) and comment the cubic part (l.152). To get the reference data for
the 2D cubic interpolator, do the contrary.
Reference interpolated data is stored in 'interp_data'.
Reference extrapolated data with nearest method is stored in 'extrap_data_nea'
Reference extrapolated data with linear method is stored in 'extrap_data_lin'
Reference extrapolated data with quadratic method is stored in 'extrap_data_qua'
"""

import numpy as np
from scipy.interpolate import interp2d
from scipy.linalg import inv
from numpy.linalg import tensorsolve

# import sys
# sys.path.append('/home/cbert/code/Raysect')
# from cherab.math.mapping.interpolators import interpolators2d

np.set_printoptions(12, 30000, linewidth=230, formatter={'float': lambda x: format(x, ' 0.12e')})  # 10 values in a line

x = np.linspace(0, 1, 10)
y = np.linspace(0, 2, 10)
x2 = x*x
x3 = x2*x
y2 = y*y
y3 = y2*y
xsamples = np.linspace(0, 1, 30)
ysamples = np.linspace(0, 2, 30)
xsamples_ex = np.array([0 - 0.08, 0 - 0.05, 0 - 0.02] + list(xsamples) + [1 + 0.02, 1 + 0.05, 1 + 0.08], dtype=np.float64)
ysamples_ex = np.array([0 - 0.16, 0 - 0.10, 0 - 0.04] + list(ysamples) + [2 + 0.04, 2 + 0.10, 2 + 0.16], dtype=np.float64)
extrapol_xdomains = [(0, 3), (3, 30 + 3), (30 + 3, 30 + 6)]
extrapol_ydomains = [(0, 3), (3, 30 + 3), (30 + 3, 30 + 6)]

deltax = x[2] - x[0]
deltay = y[2] - y[0]

def point_to_indices(a, b):
    ia = 0
    ib = 0
    if a < x[0] or a > x[-1] or b < y[0] or b > y[-1]:
        raise ValueError("point outside interpolation area")
    while not (x[ia] <= a <= x[ia+1]):
        ia += 1
    while not (y[ib] <= b <= y[ib+1]):
        ib += 1

    return ia, ib

interp_data = np.zeros((30, 30), dtype=np.float64)

extrap_data_nea = [[
                    np.zeros((3, 3), dtype=np.float64),
                    np.zeros((3, 30), dtype=np.float64),
                    np.zeros((3, 3), dtype=np.float64)
                   ],[
                    np.zeros((30, 3), dtype=np.float64),
                    None,
                    np.zeros((30, 3), dtype=np.float64)
                   ],[
                    np.zeros((3, 3), dtype=np.float64),
                    np.zeros((3, 30), dtype=np.float64),
                    np.zeros((3, 3), dtype=np.float64)
                   ]]

extrap_data_lin = [[
                    np.zeros((3, 3), dtype=np.float64),
                    np.zeros((3, 30), dtype=np.float64),
                    np.zeros((3, 3), dtype=np.float64)
                   ],[
                    np.zeros((30, 3), dtype=np.float64),
                    None,
                    np.zeros((30, 3), dtype=np.float64)
                   ],[
                    np.zeros((3, 3), dtype=np.float64),
                    np.zeros((3, 30), dtype=np.float64),
                    np.zeros((3, 3), dtype=np.float64)
                   ]]

extrap_data_qua = [[
                    np.zeros((3, 3), dtype=np.float64),
                    np.zeros((3, 30), dtype=np.float64),
                    np.zeros((3, 3), dtype=np.float64)
                   ],[
                    np.zeros((30, 3), dtype=np.float64),
                    None,
                    np.zeros((30, 3), dtype=np.float64)
                   ],[
                    np.zeros((3, 3), dtype=np.float64),
                    np.zeros((3, 30), dtype=np.float64),
                    np.zeros((3, 3), dtype=np.float64)
                   ]]

def f(a, b):
    return (a-b)*b*np.cos(10*a)

data = np.zeros((10, 10), dtype=np.float64)
for i in range(10):
    for j in range(10):
        data[i, j] = f(x[i], y[j])


# ##### linear ##### #

# interp_func = interp2d(x, y, data.T, kind='linear')
#
# for i in range(30):
#     for j in range(30):
#         interp_data[i, j] = interp_func(xsamples[i], ysamples[j])
#
# def e0(a, b):
#     a_n = max(min(a, x.max()), x.min())
#     b_n = max(min(b, y.max()), y.min())
#     return interp_func(a_n, b_n)
#
# def e1(a, b):
#     a_n = max(min(a, x.max()), x.min())
#     b_n = max(min(b, y.max()), y.min())
#     k = max(abs(a - a_n), abs(b - b_n)) * 1000
#     a_ref = a_n - (a - a_n) / k
#     b_ref = b_n - (b - b_n) / k
#
#     da = (interp_func(a_n, b_n) - interp_func(a_ref, b_n)) / (a_n - a_ref) if a_n != a_ref else 0.
#     db = (interp_func(a_n, b_n) - interp_func(a_n, b_ref)) / (b_n - b_ref) if b_n != b_ref else 0.
#     dadb = (interp_func(a_n, b_n) - interp_func(a_n, b_ref) - interp_func(a_ref, b_n) + interp_func(a_ref, b_ref)) / (a_n - a_ref) / (b_n - b_ref) if a_n != a_ref and b_n != b_ref else 0.
#
#     return interp_func(a_n, b_n) + \
#            (a - a_n) * da + \
#            (b - b_n) * db + \
#            (a - a_n) * (b - b_n) * dadb
#
# whole_calc_data_nea = np.zeros((36, 36), dtype=np.float64)
# whole_calc_data_nea[3:33, 3:33] = interp_data
# whole_calc_data_lin = np.zeros((36, 36), dtype=np.float64)
# whole_calc_data_lin[3:33, 3:33] = interp_data
#
# for nx in range(3):
#     for ny in range(3):
#         if nx != 1 or ny != 1:
#             mini, maxi = extrapol_xdomains[nx]
#             minj, maxj = extrapol_ydomains[ny]
#             for iex in range(mini, maxi):
#                 for jex in range(minj, maxj):
#                     extrap_data_nea[nx][ny][iex - mini, jex - minj] = e0(xsamples_ex[iex], ysamples_ex[jex])
#                     extrap_data_lin[nx][ny][iex - mini, jex - minj] = e1(xsamples_ex[iex], ysamples_ex[jex])
#             whole_calc_data_nea[mini:maxi, minj:maxj] = extrap_data_nea[nx][ny]
#             whole_calc_data_lin[mini:maxi, minj:maxj] = extrap_data_lin[nx][ny]

# ##### cubic ##### #

const_mat = np.zeros((9*9*16, 9*9*16), dtype=np.float64)  # ~100Mb

def elt_mat(i, j):
    return np.array([
[  x3[i  ]*y3[j  ],   x3[i  ]*y2[j  ],   x3[i  ]* y[j  ],   x3[i  ],   x2[i  ]*y3[j  ],   x2[i  ]*y2[j  ],   x2[i  ]* y[j  ],   x2[i  ],    x[i  ]*y3[j  ],    x[i  ]*y2[j  ], x[i  ]*y[j  ], x[i  ], y3[j  ], y2[j  ], y[j  ], 1.],
[  x3[i+1]*y3[j  ],   x3[i+1]*y2[j  ],   x3[i+1]* y[j  ],   x3[i+1],   x2[i+1]*y3[j  ],   x2[i+1]*y2[j  ],   x2[i+1]* y[j  ],   x2[i+1],    x[i+1]*y3[j  ],    x[i+1]*y2[j  ], x[i+1]*y[j  ], x[i+1], y3[j  ], y2[j  ], y[j  ], 1.],
[  x3[i  ]*y3[j+1],   x3[i  ]*y2[j+1],   x3[i  ]* y[j+1],   x3[i  ],   x2[i  ]*y3[j+1],   x2[i  ]*y2[j+1],   x2[i  ]* y[j+1],   x2[i  ],    x[i  ]*y3[j+1],    x[i  ]*y2[j+1], x[i  ]*y[j+1], x[i  ], y3[j+1], y2[j+1], y[j+1], 1.],
[  x3[i+1]*y3[j+1],   x3[i+1]*y2[j+1],   x3[i+1]* y[j+1],   x3[i+1],   x2[i+1]*y3[j+1],   x2[i+1]*y2[j+1],   x2[i+1]* y[j+1],   x2[i+1],    x[i+1]*y3[j+1],    x[i+1]*y2[j+1], x[i+1]*y[j+1], x[i+1], y3[j+1], y2[j+1], y[j+1], 1.],

[3*x2[i  ]*y3[j  ], 3*x2[i  ]*y2[j  ], 3*x2[i  ]* y[j  ], 3*x2[i  ], 2* x[i  ]*y3[j  ], 2* x[i  ]*y2[j  ], 2* x[i  ]* y[j  ], 2* x[i  ],   y3[j  ]        ,   y2[j  ]        , y[j  ]       , 1., 0., 0., 0., 0.],
[3*x2[i+1]*y3[j  ], 3*x2[i+1]*y2[j  ], 3*x2[i+1]* y[j  ], 3*x2[i+1], 2* x[i+1]*y3[j  ], 2* x[i+1]*y2[j  ], 2* x[i+1]* y[j  ], 2* x[i+1],   y3[j  ]        ,   y2[j  ]        , y[j  ]       , 1., 0., 0., 0., 0.],
[3*x2[i  ]*y3[j+1], 3*x2[i  ]*y2[j+1], 3*x2[i  ]* y[j+1], 3*x2[i  ], 2* x[i  ]*y3[j+1], 2* x[i  ]*y2[j+1], 2* x[i  ]* y[j+1], 2* x[i  ],   y3[j+1]        ,   y2[j+1]        , y[j+1]       , 1., 0., 0., 0., 0.],
[3*x2[i+1]*y3[j+1], 3*x2[i+1]*y2[j+1], 3*x2[i+1]* y[j+1], 3*x2[i+1], 2* x[i+1]*y3[j+1], 2* x[i+1]*y2[j+1], 2* x[i+1]* y[j+1], 2* x[i+1],   y3[j+1]        ,   y2[j+1]        , y[j+1]       , 1., 0., 0., 0., 0.],

[3*x3[i  ]*y2[j  ], 2*x3[i  ]* y[j  ],   x3[i  ]        , 0.       , 3*x2[i  ]*y2[j  ], 2*x2[i  ]* y[j  ],   x2[i  ]        , 0.       , 3* x[i  ]*y2[j  ], 2* x[i  ]* y[j  ], x[i  ]       , 0., 3*y2[j  ], 2*y[j  ], 1., 0.],
[3*x3[i+1]*y2[j  ], 2*x3[i+1]* y[j  ],   x3[i+1]        , 0.       , 3*x2[i+1]*y2[j  ], 2*x2[i+1]* y[j  ],   x2[i+1]        , 0.       , 3* x[i+1]*y2[j  ], 2* x[i+1]* y[j  ], x[i+1]       , 0., 3*y2[j  ], 2*y[j  ], 1., 0.],
[3*x3[i  ]*y2[j+1], 2*x3[i  ]* y[j+1],   x3[i  ]        , 0.       , 3*x2[i  ]*y2[j+1], 2*x2[i  ]* y[j+1],   x2[i  ]        , 0.       , 3* x[i  ]*y2[j+1], 2* x[i  ]* y[j+1], x[i  ]       , 0., 3*y2[j+1], 2*y[j+1], 1., 0.],
[3*x3[i+1]*y2[j+1], 2*x3[i+1]* y[j+1],   x3[i+1]        , 0.       , 3*x2[i+1]*y2[j+1], 2*x2[i+1]* y[j+1],   x2[i+1]        , 0.       , 3* x[i+1]*y2[j+1], 2* x[i+1]* y[j+1], x[i+1]       , 0., 3*y2[j+1], 2*y[j+1], 1., 0.],

[9*x2[i  ]*y2[j  ], 6*x2[i  ]* y[j  ], 3*x2[i  ]        , 0.       , 6* x[i  ]*y2[j  ], 4* x[i  ]* y[j  ], 2* x[i  ]        , 0.       , 3*y2[j  ]        , 2* y[j  ]        , 1.           , 0., 0., 0., 0., 0.],
[9*x2[i+1]*y2[j  ], 6*x2[i+1]* y[j  ], 3*x2[i+1]        , 0.       , 6* x[i+1]*y2[j  ], 4* x[i+1]* y[j  ], 2* x[i+1]        , 0.       , 3*y2[j  ]        , 2* y[j  ]        , 1.           , 0., 0., 0., 0., 0.],
[9*x2[i  ]*y2[j+1], 6*x2[i  ]* y[j+1], 3*x2[i  ]        , 0.       , 6* x[i  ]*y2[j+1], 4* x[i  ]* y[j+1], 2* x[i  ]        , 0.       , 3*y2[j+1]        , 2* y[j+1]        , 1.           , 0., 0., 0., 0., 0.],
[9*x2[i+1]*y2[j+1], 6*x2[i+1]* y[j+1], 3*x2[i+1]        , 0.       , 6* x[i+1]*y2[j+1], 4* x[i+1]* y[j+1], 2* x[i+1]        , 0.       , 3*y2[j+1]        , 2* y[j+1]        , 1.           , 0., 0., 0., 0., 0.]],
                    dtype=np.float64)

for i in range(9):
    for j in range(9):
        l = i + 9 * j
        const_mat[16*l:16*(l+1), 16*l:16*(l+1)] = elt_mat(i, j)

const_vec = np.zeros((9*9*16,), dtype=np.float64)

def d(i, j):
    ieff = max(min(i, 9), 0)
    jeff = max(min(j, 9), 0)
    return data[ieff, jeff]

def xf(i):
    ieff = max(min(i, 9), 0)
    return x[ieff]

def yf(j):
    jeff = max(min(j, 9), 0)
    return y[jeff]

def elt_vec(i, j):
    return np.array([
d(i+0, j+0),
d(i+1, j+0),
d(i+0, j+1),
d(i+1, j+1),

(d(i+1, j+0) - d(i-1, j+0))/(xf(i+1) - xf(i-1)),
(d(i+2, j+0) - d(i+0, j+0))/(xf(i+2) - xf(i  )),
(d(i+1, j+1) - d(i-1, j+1))/(xf(i+1) - xf(i-1)),
(d(i+2, j+1) - d(i+0, j+1))/(xf(i+2) - xf(i  )),

(d(i+0, j+1) - d(i+0, j-1))/(yf(j+1) - yf(j-1)),
(d(i+1, j+1) - d(i+1, j-1))/(yf(j+1) - yf(j-1)),
(d(i+0, j+2) - d(i+0, j+0))/(yf(j+2) - yf(j  )),
(d(i+1, j+2) - d(i+1, j+0))/(yf(j+2) - yf(j  )),

(d(i+1, j+1) - d(i+1, j-1) - d(i-1, j+1) + d(i-1, j-1))/(xf(i+1) - xf(i-1))/(yf(j+1) - yf(j-1)),
(d(i+2, j+1) - d(i+2, j-1) - d(i+0, j+1) + d(i+0, j-1))/(xf(i+2) - xf(i  ))/(yf(j+1) - yf(j-1)),
(d(i+1, j+2) - d(i+1, j+0) - d(i-1, j+2) + d(i-1, j+0))/(xf(i+1) - xf(i-1))/(yf(j+2) - yf(j  )),
(d(i+2, j+2) - d(i+2, j+0) - d(i+0, j+2) + d(i+0, j+0))/(xf(i+2) - xf(i  ))/(yf(j+2) - yf(j  )),
],
                    dtype=np.float64)

for i in range(9):
    for j in range(9):
        l = i + 9 * j
        const_vec[16*l:16*(l+1)] = elt_vec(i, j)

# inv_const_mat = inv(const_mat)
# coeffs = inv_const_mat.dot(const_vec)

coeffs = tensorsolve(const_mat, const_vec)


def P(i, j):
    if 0 <= i <= 8 and 0 <= j <= 8:
        l = 16*(i+9*j)
        return lambda a, b: coeffs[l+0 ]*a*a*a*b*b*b + \
                            coeffs[l+1 ]*a*a*a*b*b + \
                            coeffs[l+2 ]*a*a*a*b + \
                            coeffs[l+3 ]*a*a*a + \
                            coeffs[l+4 ]*a*a*b*b*b + \
                            coeffs[l+5 ]*a*a*b*b + \
                            coeffs[l+6 ]*a*a*b + \
                            coeffs[l+7 ]*a*a + \
                            coeffs[l+8 ]*a*b*b*b + \
                            coeffs[l+9 ]*a*b*b + \
                            coeffs[l+10]*a*b + \
                            coeffs[l+11]*a + \
                            coeffs[l+12]*b*b*b + \
                            coeffs[l+13]*b*b + \
                            coeffs[l+14]*b + \
                            coeffs[l+15]
    else:
        raise ValueError("i and j must be between 0 and 8 included")

def dP_dx(i, j):
    if 0 <= i <= 8 and 0 <= j <= 8:
        l = 16*(i+9*j)
        return lambda a, b: 3*coeffs[l+0 ]*a*a*b*b*b + \
                            3*coeffs[l+1 ]*a*a*b*b + \
                            3*coeffs[l+2 ]*a*a*b + \
                            3*coeffs[l+3 ]*a*a + \
                            2*coeffs[l+4 ]*a*b*b*b + \
                            2*coeffs[l+5 ]*a*b*b + \
                            2*coeffs[l+6 ]*a*b + \
                            2*coeffs[l+7 ]*a + \
                            coeffs[l+8 ]*b*b*b + \
                            coeffs[l+9 ]*b*b + \
                            coeffs[l+10]*b + \
                            coeffs[l+11]
    else:
        raise ValueError("i and j must be between 0 and 8 included")

def d2P_dx2(i, j):
    if 0 <= i <= 8 and 0 <= j <= 8:
        l = 16*(i+9*j)
        return lambda a, b: 6*coeffs[l+0 ]*a*b*b*b + \
                            6*coeffs[l+1 ]*a*b*b + \
                            6*coeffs[l+2 ]*a*b + \
                            6*coeffs[l+3 ]*a + \
                            2*coeffs[l+4 ]*b*b*b + \
                            2*coeffs[l+5 ]*b*b + \
                            2*coeffs[l+6 ]*b + \
                            2*coeffs[l+7 ]
    else:
        raise ValueError("i and j must be between 0 and 8 included")

def dP_dy(i, j):
    if 0 <= i <= 8 and 0 <= j <= 8:
        l = 16*(i+9*j)
        return lambda a, b: 3*coeffs[l+0 ]*a*a*a*b*b + \
                            2*coeffs[l+1 ]*a*a*a*b + \
                            coeffs[l+2 ]*a*a*a + \
                            3*coeffs[l+4 ]*a*a*b*b + \
                            2*coeffs[l+5 ]*a*a*b + \
                            coeffs[l+6 ]*a*a + \
                            3*coeffs[l+8 ]*a*b*b + \
                            2*coeffs[l+9 ]*a*b + \
                            coeffs[l+10]*a + \
                            3*coeffs[l+12]*b*b + \
                            2*coeffs[l+13]*b + \
                            coeffs[l+14]
    else:
        raise ValueError("i and j must be between 0 and 8 included")

def d2P_dy2(i, j):
    if 0 <= i <= 8 and 0 <= j <= 8:
        l = 16*(i+9*j)
        return lambda a, b: 6*coeffs[l+0 ]*a*a*a*b + \
                            2*coeffs[l+1 ]*a*a*a + \
                            6*coeffs[l+4 ]*a*a*b + \
                            2*coeffs[l+5 ]*a*a + \
                            6*coeffs[l+8 ]*a*b + \
                            2*coeffs[l+9 ]*a + \
                            6*coeffs[l+12]*b + \
                            2*coeffs[l+13]
    else:
        raise ValueError("i and j must be between 0 and 8 included")

def d2P_dxdy(i, j):
    if 0 <= i <= 8 and 0 <= j <= 8:
        l = 16*(i+9*j)
        return lambda a, b: 9*coeffs[l+0 ]*a*a*b*b + \
                            6*coeffs[l+1 ]*a*a*b + \
                            3*coeffs[l+2 ]*a*a + \
                            6*coeffs[l+4 ]*a*b*b + \
                            4*coeffs[l+5 ]*a*b + \
                            2*coeffs[l+6 ]*a + \
                            3*coeffs[l+8 ]*b*b + \
                            2*coeffs[l+9 ]*b + \
                            coeffs[l+10]
    else:
        raise ValueError("i and j must be between 0 and 8 included")

def interp_func(a, b):
    return P(*point_to_indices(a, b))(a, b)

for i in range(30):
    for j in range(30):
        interp_data[i, j] = interp_func(xsamples[i], ysamples[j])

# compare_func = interpolators2d.Interpolate2DCubic(x, y, data)
# compare_data = np.zeros((30, 30), dtype=np.float64)
# for i in range(30):
#     for j in range(30):
#         compare_data[i, j] = compare_func(xsamples[i], ysamples[j])


def e0(a, b):
    a_n = max(min(a, x.max()), x.min())
    b_n = max(min(b, y.max()), y.min())
    return interp_func(a_n, b_n)

def e1(a, b):
    a_n = max(min(a, x.max()), x.min())
    b_n = max(min(b, y.max()), y.min())
    i, j = point_to_indices(a_n, b_n)

    return P(i, j)(a_n, b_n) + \
           (a - a_n) * dP_dx(i, j)(a_n, b_n) + \
           (b - b_n) * dP_dy(i, j)(a_n, b_n)

def e2(a, b):
    a_n = max(min(a, x.max()), x.min())
    b_n = max(min(b, y.max()), y.min())
    i, j = point_to_indices(a_n, b_n)

    return e1(a, b) + \
           (a - a_n) * (b - b_n) * d2P_dxdy(i, j)(a_n, b_n) + \
           0.5 * (a - a_n)**2 * d2P_dx2(i, j)(a_n, b_n) + \
           0.5 * (b - b_n)**2 * d2P_dy2(i, j)(a_n, b_n)



whole_calc_data_nea = np.zeros((36, 36), dtype=np.float64)
whole_calc_data_nea[3:33, 3:33] = interp_data
whole_calc_data_lin = np.zeros((36, 36), dtype=np.float64)
whole_calc_data_lin[3:33, 3:33] = interp_data
whole_calc_data_qua = np.zeros((36, 36), dtype=np.float64)
whole_calc_data_qua[3:33, 3:33] = interp_data

for nx in range(3):
    for ny in range(3):
        if nx != 1 or ny != 1:
            mini, maxi = extrapol_xdomains[nx]
            minj, maxj = extrapol_ydomains[ny]
            for iex in range(mini, maxi):
                for jex in range(minj, maxj):
                    extrap_data_nea[nx][ny][iex - mini, jex - minj] = e0(xsamples_ex[iex], ysamples_ex[jex])
                    extrap_data_lin[nx][ny][iex - mini, jex - minj] = e1(xsamples_ex[iex], ysamples_ex[jex])
                    extrap_data_qua[nx][ny][iex - mini, jex - minj] = e2(xsamples_ex[iex], ysamples_ex[jex])
            whole_calc_data_nea[mini:maxi, minj:maxj] = extrap_data_nea[nx][ny]
            whole_calc_data_lin[mini:maxi, minj:maxj] = extrap_data_lin[nx][ny]
            whole_calc_data_qua[mini:maxi, minj:maxj] = extrap_data_qua[nx][ny]




# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.


import numpy as np
import matplotlib.pyplot as plt
from scipy.constants import electron_mass, atomic_mass

from raysect.core.math.function.float import Arg2D, Exp2D, Sqrt2D
from raysect.primitive import Cylinder
from raysect.optical import World, translate, Point3D, Vector3D, rotate_basis, Spectrum
from raysect.optical.observer import PinholeCamera, PowerPipeline2D

from cherab.core import Species, Maxwellian, Plasma, Line
from cherab.core.math import sample3d, AxisymmetricMapper
from cherab.core.atomic import deuterium
from cherab.core.model import ExcitationLine
from cherab.openadas import OpenADAS


def NeutralFunction(peak_value, sigma, magnetic_axis, lcfs_radius=1):
    """A neutral profile that is constant outside the plasma,
       then exponentially decays inside the LCFS."""
    raxis = magnetic_axis[0]
    zaxis = magnetic_axis[1]
    radius_from_axis = Sqrt2D((Arg2D('x') - raxis)**2 + (Arg2D('y') - zaxis)**2)
    scale = Exp2D(-((radius_from_axis - lcfs_radius)**2) / (2 * sigma**2))
    inside_lcfs = (radius_from_axis <= lcfs_radius)
    # density = peak * scale * inside_lcfs + peak * (inside_lcfs - 1).
    # Rearrange so inside_lcfs and scale are only called once each.
    density = peak_value * (inside_lcfs * (scale - 1) + 1)
    return AxisymmetricMapper(density)


def IonFunction(v_core, v_lcfs, magnetic_axis, p=4, q=3, lcfs_radius=1):
    """An approximate toroidal plasma profile that follows a double
       quadratic between the LCFS and magnetic axis."""
    r_axis = magnetic_axis[0]
    z_axis = magnetic_axis[1]
    radius_from_axis = Sqrt2D((Arg2D('x') - r_axis)**2 + (Arg2D('y') - z_axis)**2)
    density = (v_core - v_lcfs) * (1 - (radius_from_axis / lcfs_radius)**p)**q + v_lcfs
    inside_lcfs = (radius_from_axis <= lcfs_radius)
    return AxisymmetricMapper(density * inside_lcfs)


# tunables
peak_density = 1e19
peak_temperature = 2500
magnetic_axis = (2.5, 0)


# setup scenegraph
world = World()


###################
# plasma creation #

plasma = Plasma(parent=world)
plasma.atomic_data = OpenADAS(permit_extrapolation=True)
plasma.geometry = Cylinder(3.5, 2.2, transform=translate(0, 0, -1.1))
plasma.geometry_transform = translate(0, 0, -1.1)

# No net velocity for any species
zero_velocity = Vector3D(0, 0, 0)

# define neutral species distribution
d0_density = NeutralFunction(peak_density, 0.1, magnetic_axis)
d0_temperature = 0.5  # constant 0.5eV temperature for all neutrals
d0_distribution = Maxwellian(d0_density, d0_temperature, zero_velocity,
                             deuterium.atomic_weight * atomic_mass)
d0_species = Species(deuterium, 0, d0_distribution)

# define deuterium ion species distribution
d1_density = IonFunction(peak_density, 0, magnetic_axis)
d1_temperature = IonFunction(peak_temperature, 0, magnetic_axis)
d1_distribution = Maxwellian(d1_density, d1_temperature, zero_velocity,
                             deuterium.atomic_weight * atomic_mass)
d1_species = Species(deuterium, 1, d1_distribution)

# define the electron distribution
e_density = IonFunction(peak_density, 0, magnetic_axis)
e_temperature = IonFunction(peak_temperature, 0, magnetic_axis)
e_distribution = Maxwellian(e_density, e_temperature, zero_velocity, electron_mass)

# define species
plasma.b_field = Vector3D(1.0, 1.0, 1.0)
plasma.electron_distribution = e_distribution
plasma.composition = [d0_species, d1_species]

# Add a balmer alpha line for visualisation purposes
d_alpha_excit = ExcitationLine(Line(deuterium, 0, (3, 2)))
plasma.models = [d_alpha_excit]


####################
# Visualise Plasma #

# Run some plots to check the distribution functions and emission profile are as expected
r, _, z, t_samples = sample3d(d1_temperature, (0, 4, 200), (0, 0, 1), (-2, 2, 200))
plt.imshow(np.transpose(np.squeeze(t_samples)), extent=[0, 4, -2, 2])
plt.colorbar()
plt.axis('equal')
plt.xlabel('r axis')
plt.ylabel('z axis')
plt.title("Ion temperature profile in r-z plane")

plt.figure()
r, _, z, t_samples = sample3d(d1_temperature, (-4, 4, 400), (-4, 4, 400), (0, 0, 1))
plt.imshow(np.transpose(np.squeeze(t_samples)), extent=[-4, 4, -4, 4])
plt.colorbar()
plt.axis('equal')
plt.xlabel('x axis')
plt.ylabel('y axis')
plt.title("Ion temperature profile in x-y plane")

plt.figure()
r, _, z, t_samples = sample3d(d0_density, (0, 4, 200), (0, 0, 1), (-2, 2, 200))
plt.imshow(np.transpose(np.squeeze(t_samples)), extent=[0, 4, -2, 2])
plt.colorbar()
plt.axis('equal')
plt.xlabel('r axis')
plt.ylabel('z axis')
plt.title("Neutral Density profile in r-z plane")

plt.figure()
xrange = np.linspace(0, 4, 200)
yrange = np.linspace(-2, 2, 200)
d_alpha_rz_intensity = np.zeros((200, 200))
direction = Vector3D(0, 1, 0)
for i, x in enumerate(xrange):
    for j, y in enumerate(yrange):
        emission = d_alpha_excit.emission(Point3D(x, 0.0, y), direction, Spectrum(650, 660, 1))
        d_alpha_rz_intensity[j, i] = emission.total()
plt.imshow(d_alpha_rz_intensity, extent=[0, 4, -2, 2], origin='lower')
plt.colorbar()
plt.xlabel('r axis')
plt.ylabel('z axis')
plt.title("D-alpha emission in R-Z")


camera = PinholeCamera((256, 256), pipelines=[PowerPipeline2D()], parent=world)
camera.transform = translate(2.5, -4.5, 0)*rotate_basis(Vector3D(0, 1, 0), Vector3D(0, 0, 1))
camera.pixel_samples = 1

plt.ion()
camera.observe()
plt.ioff()
plt.show()



import unittest
import numpy as np
from cherab.core.math.caching import Caching3D


class TestCaching3D(unittest.TestCase):

    def setUp(self):
        self.function = lambda x, y, z: (x - y) * y * np.cos(10 * x) * np.exp(z) + z
        self.grad_f = lambda x, y, z: np.array([y * np.exp(z) * (np.cos(10 * x) - 10 * (x - y) * np.sin(10 * x)),
                                                (x - 2 * y) * np.cos(10 * x) * np.exp(z),
                                                (x - y) * y * np.cos(10 * x) * np.exp(z) + 1])
        self.hess_f = lambda x, y, z: np.array([[10 * y * (-2 * np.sin(10 * x) - 10 * (x - y) * np.cos(10 * x)) * np.exp(z), np.cos(10 * x) - 10 * (x - 2 * y) * np.sin(10 * x) * np.exp(z), y * np.exp(z) * (np.cos(10 * x) - 10 * (x - y) * np.sin(10 * x))],
                                                [np.cos(10 * x) - 10 * (x - 2 * y) * np.sin(10 * x) * np.exp(z)            , -2 * np.cos(10 * x) * np.exp(z)                               , (x - 2 * y) * np.cos(10 * x) * np.exp(z)                        ],
                                                [y * np.exp(z) * (np.cos(10 * x) - 10 * (x - y) * np.sin(10 * x))          , (x - 2 * y) * np.cos(10 * x) * np.exp(z)                      , (x - y) * y * np.cos(10 * x) * np.exp(z)                        ]])
        xmin, xmax = -5, 2
        ymin, ymax = 0.5, 4
        zmin, zmax = -8.3, -1.1
        self.space_area = xmin, xmax, ymin, ymax, zmin, zmax
        xresolution = 0.1
        yresolution = 0.05
        zresolution = 0.2
        self.resolution = xresolution, yresolution, zresolution

    def tearDown(self):
        del self.function

    def tolerance(self, x, y, z):
        np_res = np.array(self.resolution)
        tolerance = (self.grad_f(x, y, z) * np_res).sum() + 0.5 * (self.hess_f(x, y, z).dot(np_res) * np_res).sum()
        return max(abs(tolerance), 0.2)

    def test_values(self):
        cached_func = Caching3D(self.function, self.space_area, self.resolution)

        for x in np.linspace(self.space_area[0], self.space_area[1], 20):
            for y in np.linspace(self.space_area[2], self.space_area[3], 20):
                for z in np.linspace(self.space_area[4], self.space_area[5], 20):
                    self.assertAlmostEqual(cached_func(x, y, z), self.function(x, y, z), delta=self.tolerance(x, y, z),
                                           msg='Cached function at ({}, {}, {}) is too far from exact function!'.format(x, y, z))
                    self.assertAlmostEqual(cached_func(x, y, z), self.function(x, y, z), delta=1.,
                                           msg='Cached function at ({}, {}, {}) is too far from exact function!'.format(x, y, z))


if __name__ == '__main__':
    unittest.main()


# Copyright 2016-2023 Euratom
# Copyright 2016-2023 United Kingdom Atomic Energy Authority
# Copyright 2016-2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

import unittest

import numpy as np

from raysect.core import World, Vector3D, translate

from cherab.core import Beam
from cherab.core.atomic import AtomicData, BeamStoppingRate
from cherab.core.atomic import deuterium
from cherab.tools.plasmas.slab import build_constant_slab_plasma
from cherab.core.model import SingleRayAttenuator

from cherab.core.utility import EvAmuToMS, EvToJ


class ConstantBeamStoppingRate(BeamStoppingRate):
    """
    Constant beam CX PEC for test purpose.
    """

    def __init__(self, donor_metastable, value):
        self.donor_metastable = donor_metastable
        self.value = value

    def evaluate(self, energy, density, temperature):

        return self.value


class MockAtomicData(AtomicData):
    """Fake atomic data for test purpose."""

    def beam_stopping_rate(self, beam_ion, plasma_ion, charge):

        return ConstantBeamStoppingRate(1, 1.e-13)


class TestBeam(unittest.TestCase):

    def setUp(self):

        self.atomic_data = MockAtomicData()

        self.world = World()

        self.plasma_density = 1.e19
        self.plasma_temperature = 1.e3
        plasma_species = [(deuterium, 1, self.plasma_density, self.plasma_temperature, Vector3D(0, 0, 0))]
        plasma = build_constant_slab_plasma(length=1, width=1, height=1,
                                            electron_density=self.plasma_density,
                                            electron_temperature=self.plasma_temperature,
                                            plasma_species=plasma_species)
        plasma.atomic_data = self.atomic_data
        plasma.parent = self.world

        beam = Beam(transform=translate(0.5, 0, 0))
        beam.atomic_data = self.atomic_data
        beam.plasma = plasma
        beam.attenuator = SingleRayAttenuator(clamp_to_zero=True)
        beam.energy = 50000
        beam.power = 1e6
        beam.temperature = 10
        beam.element = deuterium
        beam.parent = self.world
        beam.sigma = 0.2
        beam.divergence_x = 1.
        beam.divergence_y = 2.
        beam.length = 10.

        self.plasma = plasma
        self.beam = beam

    def test_beam_density(self):

        z0 = 0.8
        x0, y0 = 0.5, 0.5

        density_on_axis = self.beam.density(0, 0, z0)
        density_off_axis = self.beam.density(x0, y0, z0)
        density_outside_beam = self.beam.density(0, 0, -1)

        # validating

        speed = EvAmuToMS.to(self.beam.energy)
        # constant stopping rate
        stopping_rate = self.atomic_data.beam_stopping_rate(deuterium, deuterium, 1)(0, 0, 0)
        attenuation_factor = np.exp(-z0 * self.plasma_density * stopping_rate / speed)

        beam_particle_rate = self.beam.power / EvToJ.to(self.beam.energy * deuterium.atomic_weight)

        sigma0_sqr = self.beam.sigma**2
        tanxdiv = np.tan(np.deg2rad(self.beam.divergence_x))
        tanydiv = np.tan(np.deg2rad(self.beam.divergence_y))
        sigma_x = np.sqrt(sigma0_sqr + (z0 * tanxdiv)**2)
        sigma_y = np.sqrt(sigma0_sqr + (z0 * tanydiv)**2)

        norm_radius_sqr = ((x0 / sigma_x)**2 + (y0 / sigma_y)**2)

        gaussian_sample_on_axis = 1. / (2 * np.pi * sigma_x * sigma_y)
        gaussian_sample_off_axis = np.exp(-0.5 * norm_radius_sqr) / (2 * np.pi * sigma_x * sigma_y)

        test_density_on_axis = beam_particle_rate / speed * gaussian_sample_on_axis * attenuation_factor
        test_density_off_axis = beam_particle_rate / speed * gaussian_sample_off_axis * attenuation_factor

        self.assertAlmostEqual(density_on_axis / test_density_on_axis, 1., delta=1.e-12,
                               msg='Beam.density() gives a wrong value on the beam axis.')
        self.assertAlmostEqual(density_off_axis / test_density_off_axis, 1., delta=1.e-12,
                               msg='Beam.density() gives a wrong value off the beam axis.')
        self.assertEqual(density_outside_beam, 0,
                         msg='Beam.density() gives a non-zero value outside beam.')

    def test_beam_direction(self):
        # setting up the model

        z0 = 0.8
        x0, y0 = 0.5, 0.5

        direction_on_axis = self.beam.direction(0, 0, z0)
        direction_off_axis = self.beam.direction(x0, y0, z0)
        direction_outside_beam = self.beam.direction(0, 0, -1)

        # validating

        sigma0_sqr = self.beam.sigma**2
        z_tanx_sqr = (z0 * np.tan(np.deg2rad(self.beam.divergence_x)))**2
        z_tany_sqr = (z0 * np.tan(np.deg2rad(self.beam.divergence_y)))**2

        ex = x0 * z_tanx_sqr / (sigma0_sqr + z_tanx_sqr)
        ey = y0 * z_tany_sqr / (sigma0_sqr + z_tany_sqr)
        ez = z0

        test_direction_off_axis = Vector3D(ex, ey, ez).normalise()

        self.assertEqual(direction_on_axis, Vector3D(0, 0, 1),
                         msg='Beam.density() gives a wrong value on the beam axis.')
        for v, test_v in zip(direction_off_axis, test_direction_off_axis):
            self.assertAlmostEqual(v, test_v, delta=1.e-12,
                                   msg='Beam.direction() gives a wrong value off the beam axis.')
        self.assertEqual(direction_outside_beam, Vector3D(0, 0, 1),
                         msg='Beam.density() gives a non-zero value outside beam.')


if __name__ == '__main__':
    unittest.main()
